/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 * 
 * $Id: aodv.c,v 1.1 2007/06/08 09:51:39 bosterm Exp $
 */

#ifdef __BTNODE3__

#include <sys/event.h>
#include <sys/timer.h>
#include <cc/ccc.h>
#include <string.h>
#include <led/btn-led.h>
#include <stdio.h>
#include <cc/aodv.h>

/////////////////////////////////////////////////////////////////////////////////////
// Defines
/////////////////////////////////////////////////////////////////////////////////////

/**
* Packet Types
 */
#define CONTROL_TYPE    0xF0
#define RREQ_TYPE       0x01
#define RREP_TYPE       0x02
#define RREP_ACK_TYPE   0x03
#define RERR_TYPE       0x04

/**
* Flags
 */
//RREQ Flags
#define FLAG_JOIN                    128 //10000000
#define FLAG_REPAIR                   64 //01000000
#define FLAG_GRATIOUS                 32 //00100000
#define FLAG_DESTINATION_ONLY         16 //00010000
#define FLAG_UNKNOWN_SEQUENCE_NUMBER   8 //00001000

//RREP Flags
#define FLAG_REP                     128 //10000000
#define FLAG_ACK                      64 //01000000

//Routing Table Flags
#define FLAG_VALID                     8 //00001000
#define FLAG_ACTIVE                   16 //00010000

/**
* States
 **/
// Buffer states
#define FULL           0
#define FREE           1


/**
* Helpers
 */
#define MAX(a,b)        (a) < (b) ? (b) : (a)                     //Signed max()
#define UMAX(a,b)       (signed)(b) - (signed)(a) < 0 ? (a) : (b) //Unsigned max()


/////////////////////////////////////////////////////////////////////////////////////
// Defines
/////////////////////////////////////////////////////////////////////////////////////

/************************************************************************************
* Routing table structure
************************************************************************************/

/**
* Routing Table Entry
 * 
 * IMPROVE - Adding a precursor list, as proposed in Request for Comments 3561 could
 * improve performance, because of better local repair.
 */
typedef struct routingtable_entry_st {
    u_short dst_address;      ///< destination address
    u_short dsn;              ///< destination sequence number
    u_short flags;            ///< flags: Valid DSN, Active
    u_short metric;           ///< metric: lower = better. IMPROVE - A better metric than hopcounts could be used.
    u_short next_hop_address; ///< adress of next hop on path to destination
    u_long lifetime;          ///< time, when the route expires.
} __attribute__((__packed__)) routingtable_entry_t;


/************************************************************************************
* Control packet structures
************************************************************************************/

/**
* Structure of a RREQ packet
 */
typedef struct aodv_control_rreq_packet_st {
    u_char type;                 ///< Packet Type. Always MUST be set to RREQ_TYPE.
    u_char flags;                ///< Flags (5 bit) + reserved (3 bit).
    u_char hopcount;             ///< Number of hops that RREQ made.
    u_char reserved;             ///< Reserved, not present in original AODV.
    u_short id;                  ///< RREQ id (16 bit, not 32 bit as in rfc3561).
    u_short destination_address; ///< Address of the destination node (16 bit, not 32 bit as in rfc3561).
    u_short dsn;                 ///< Last known sequence number of the destination (16 bit, not 32 bit as in rfc3561).
    u_short originator_address;  ///< Originatot address (16 bit, not 32 bit as in rfc3561).
    u_short osn;                 ///< Originator sequence number (16 bit, not 32 bit as in rfc3561).
    u_short ttl;                 ///< Time to life. Number of hops remaining until packet expires.
} __attribute__((__packed__)) aodv_control_rreq_packet_t;

/**
* Structure of a RREP packet
 */
typedef struct aodv_control_rrep_packet_st {
    u_char type;                 ///< Packet Type. Always MUST be set to RREP_TYPE.
    u_char flags;                ///< Flags (2 bit) + reserved (6 bit).
    u_char prefix_sz;            ///< Reserved (3 bit) + Prefix SZ (5 bit).
    u_char hopcount;             ///< Number of hops that RREP made.
    u_short destination_address; ///< Address of the destination node (16 bit, not 32 bit as in rfc3561).
    u_short dsn;                 ///< Last known sequence number of the destination (16 bit, not 32 bit as in rfc3561).
    u_short originator_address;  ///< Originator address (16 bit, not 32 bit as in rfc3561).
    u_long lifetime;             ///< Lifetime of the route.
    u_short ttl;                 ///< Time to life. Number of hops remaining until packet expires.
} __attribute__((__packed__)) aodv_control_rrep_packet_t;

/**
* Structure of an RERR packet
 */
typedef struct aodv_control_rerr_packet_st {
    u_char type;                             ///< Packet Type. Always MUST be set to RERR_TYPE.
    u_short flags;                           ///< Flags (1 bit) + reserved (15 bit).
    u_char destcount;                        ///< Number of unreachable destiantions.
    u_short unreachable_destination_address; ///< Address of the unreachable destination (16 bit, not 32 bit as in rfc3561).
    u_short unreachable_dsn;                 ///< Last known sequence number of the unreachable destination
    u_short additional_unreachable_destination_address; ///< Additional unreachable destination addresses (if needed)
    u_short additional_unreachable_dsn;      ///< Last known additional unreachable destination sequence number.
} __attribute__((__packed__)) aodv_control_rerr_packet_t;



/**
* Structure of an RREP-ACK packet
 */
typedef struct aodv_control_rrep_ack_packet_st {
    u_char type;     ///< Packet Type. Always MUST be set to RREP_ACK_TYPE.
    u_char reserved; ///< Reserved (8 bit).
} __attribute__((__packed__)) aodv_control_rrep_ack_packet_t;


/************************************************************************************
* Buffer structures
************************************************************************************/

/**
* RREQ buffer structure
 */
typedef struct aodv_rreq_buffer_st {
    u_short id;                 ///< RREQ id.
    u_short originator_address; ///< Address of the originator of the RREQ
    u_long timestamp;           ///< Timestamp, when the entry was created.
} __attribute__((__packed__)) aodv_rreq_buffer_t;


/**
* Recv Control Buffer structure
 **/
typedef struct aodv_recv_control_buffer_st {
    ccc_packet_t *pkt;           ///< Pointer to the control packet
    u_short src;                 ///< Sender of the packet (previous hop)
    u_char state;                ///< State of this slot: FREE or FULL
} __attribute__((__packed__)) aodv_recv_control_buffer_t;

/**
* Route discovery buffer structure
 */
typedef struct aodv_discovery_buffer_st {
    aodv_control_rreq_packet_t *rreq; ///< Pointer to the RREQ packet
    u_short retry;                    ///< Counts how many time sending has been retried.
    long nextshd;                     ///< Time, when the packet should be reprocessed.
}  __attribute__((__packed__)) aodv_discovery_buffer_t;

/**
* RREP buffer structure
 */
typedef struct aodv_rrep_buffer_st {
    aodv_control_rrep_packet_t *rrep; ///< Pointer to the RREP packet
    u_short dst;                      ///< Destination (next hop) where the packet is sent to
    u_short retry;                    ///< Counts how many times sending has been retried
    long nextshd;                     ///< Time, when the packet should be reprocessed.
}  __attribute__((__packed__)) aodv_rrep_buffer_t;

/**
* Blacklist structure
 */
typedef struct aodv_blacklist_st {
    u_short address;
    u_long lifetime;
} __attribute__((__packed__)) aodv_blacklist_t;

/////////////////////////////////////////////////////////////////////////////////////
// Local variables
/////////////////////////////////////////////////////////////////////////////////////

//Routing Table
routingtable_entry_t routing_table[ROUTING_TABLE_SIZE];

//Packets
ccc_packet_t    *control_packet;

//Queues & queue control
aodv_recv_control_buffer_t  aodv_recv_control_packet_buffer[CONTROL_PACKET_QUEUE_SIZE]; //Buffer for received packets
u_short                     aodv_recv_control_packet_buffer_counter;                    //Number of packets in recv buffer

ccc_packet_t                *send_control_packet_queue[CONTROL_PACKET_QUEUE_SIZE];      //Buffer for packets to be sent
u_short                     send_control_packet_queue_counter;                          //Number of packets in send buffer

aodv_discovery_buffer_t     aodv_discovery_buffer[DISCOVERY_BUFFER_SIZE];               //Buffer for RREQs on discovery
u_short                     aodv_discovery_buffer_count;                                //Number of packets in d. buffer

aodv_rrep_buffer_t          aodv_rrep_buffer[RREP_BUFFER_SIZE];                     //Buffer for RREPs waiting for an ACK
u_short                     aodv_rrep_buffer_count;                                 //Number of pending RREPs in buffer

//Handles
HANDLE event_control_packet_recvd;                  //Signals that a control packet was received
HANDLE event_discover_request;                      //Signals that a route discovery was requested
HANDLE event_rreq_pending;                          //Signals that a RREQ is waiting for processing
HANDLE event_rrep_pending;                          //Signals that a RREP is waiting for processing

//Buffers
aodv_rreq_buffer_t   rreq_buffer[RREQ_BUFFER_SIZE]; //A buffer for RREQs that already have been seen. (avoid cycles)
aodv_blacklist_t     blacklist[BLACKLIST_SIZE];     //Blacklist to deal with unidirectional links.

//AODV Control Data
u_short aodv_sequence_no;                           //Sequence number of the node
u_short aodv_rreq_id;                               //rreq id of the current RREQ

//Mutexes
MUTEX mutex_routing_table;
MUTEX mutex_rreq_buffer;
MUTEX mutex_discovery_buffer;
MUTEX mutex_discovery_buffer_count;
MUTEX mutex_rrep_buffer;
MUTEX mutex_rrep_buffer_count;

//Statistics
u_short aodv_rreq_recv_count;              ///< Number of RREQ packets targeted for this node. (including failed)
u_short aodv_rreq_send_count;              ///< Number of RREQ packets sent from this node. (including failed)
u_short aodv_rreq_recv_fail_count;         ///< Number of failed RREQ packets passing through this node.
u_short aodv_rreq_send_fail_count;         ///< Number of send failures of RREQ packets.
u_short aodv_rreq_ignore_count;            ///< Number of ignored RREQs (received twice or more).
u_short aodv_rrep_recv_count;              ///< Number of RREP packets targeted for this node.(including failed)
u_short aodv_rrep_send_count;              ///< Number of RREP packets sent from this node. (including failed)
u_short aodv_rrep_recv_fail_count;         ///< Number of failed RREP packets passing through this node.
u_short aodv_rrep_send_fail_count;         ///< Number of send failures of RREP packets.
u_short aodv_rrep_ack_recv_count;          ///< Number of RREP-ACK packets targeted for this node.
u_short aodv_rrep_ack_send_count;          ///< Number of RREP-ACK packets sent from this node. (including failed)
u_short aodv_rrep_ack_send_fail_count;     ///< Number of send failures of RREP-ACK packets.
u_short aodv_rerr_recv_count;              ///< Number of RERR packets targeted for this node. 
u_short aodv_rerr_send_count;              ///< Number of RERR packets sent from this node. (including failed)
u_short aodv_rerr_send_fail_count;         ///< Number of send failures of RERR packets.
u_short aodv_invalid_count;                ///< Number of invalid packets targeted for this node.


//Misc
u_short                   aodv_address;           //Address of the node
u_char                    use_ack;                //If set, acks for RREP's of this node are enabled 
                                                  //IMPROVE - Sensing the link quality at runtime and enabling ACKs 
                                                  //          dinamically for each next hop would be nice. 
long                      last_broadcast;         //Timestamp when last broadcast occured, used for HELLO messages
routing_message_handler_t aodv_message_handler;   //Handler of RoutingForChipcon for aodv messages
static u_char             aodv_use_led = 0;       //Set to 1 if LED flashing is desired.

/////////////////////////////////////////////////////////////////////////////////////
// Local Function Predefinitions
/////////////////////////////////////////////////////////////////////////////////////

// Routing Table Management
inline static u_char rt_check_entry_validity(u_short i);
inline static u_char rt_check_entry_activity(u_short i);
inline static short rt_create_entry(u_short dst, u_short dsn, u_short flags, 
                                    u_short metric, u_short next_hop, int lifetime);
inline static char rt_update_entry(short handle);
inline static void rt_deactivate_next_hop(u_short next_hop); //Deactivates any routing table entry that uses next_hop
inline static short rt_get_entry(u_short dst);
inline static u_short rt_get_flags(u_short entry, u_short flags);
inline static u_short rt_get_sequence_no(u_short entry);

// Control Packet Management
inline static void *new_aodv_control_packet(int type);
static int aodv_send_control_rreq_packet(u_short addr, aodv_control_rreq_packet_t *rreq);
static int aodv_send_control_rrep_packet(u_short addr, aodv_control_rrep_packet_t *rrep);
static int aodv_send_control_rrep_ack_packet(u_short addr, aodv_control_rrep_ack_packet_t *rrep_ack);
static int aodv_send_control_rerr_packet(u_short addr, aodv_control_rerr_packet_t *rerr);
static void control_handler(ccc_packet_t *pkt);
inline static void aodv_pkt_serialize_rreq(u_char *to, aodv_control_rreq_packet_t *from);
inline static void aodv_pkt_deserialize_rreq(aodv_control_rreq_packet_t *to, u_char *from);
inline static void aodv_pkt_serialize_rrep(u_char *to, aodv_control_rrep_packet_t *from);
inline static void aodv_pkt_deserialize_rrep(aodv_control_rrep_packet_t *to, u_char *from);
inline static void aodv_pkt_serialize_rrep_ack(u_char *to, aodv_control_rrep_ack_packet_t *from);
inline static void aodv_pkt_deserialize_rrep_ack(aodv_control_rrep_ack_packet_t *to, u_char *from);
inline static void aodv_pkt_serialize_rerr(u_char *to, aodv_control_rerr_packet_t *from);
inline static void aodv_pkt_deserialize_rerr(aodv_control_rerr_packet_t *to, u_char *from);

// Buffers
static char rreq_buffer_add(u_short id, u_short originator_address);
static char rreq_buffer_lookup(u_short id, u_short originator_address);
static char rrep_buffer_add(aodv_control_rrep_packet_t *rrep, u_short dst);
static char rrep_buffer_del(u_short dst);
static inline char rreq_discovery_buffer_add(aodv_control_rreq_packet_t *rreq);
static inline char aodv_blacklist_add(u_short address);
static inline u_short aodv_blacklist_lookup(u_short address);

//Route Discovery (Main Functionallity)
static inline u_char aodv_check_discovery_in_progress(u_short dst);

//Route Maintenance Functions
//None needed, since RT uses lazy evaluation and minor updates (flags, ...) are done 
//in code.


/////////////////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////////////////

// calculates a waiting time with [ 1 .. 1 << tries  ] * slotduration
u_short aodv_beb(u_short slotduration, u_char next_try){
    if (next_try > 15) next_try = 15;
    u_short t_random = (rand() & ((1 << next_try) - 1) ) * slotduration;
    return slotduration + t_random;
}


/************************************************************************************
* Route discovery
************************************************************************************/
/**
* Starts a route discovery for a given destionation if
 * there is not already one in progress.
 *
 * @param dst [in] destination address
 * @return DISCOVERY_STARTED if a discovery is started,
 *         DISCOVERY_IN_PROGRESS if a discovery is already in progress,
 *         ROUTE_AVAILABLE if there is no need for a discovery (and none has been started).
 *
 */
u_short aodv_discover(u_short *dst)
{
    u_short i; 
    aodv_control_rreq_packet_t *rreq; 
    
    // Look up destination in Table
    i = rt_get_entry(*dst);
    
    if(i != -1 && rt_check_entry_validity(i) && rt_check_entry_activity(i)) {
        //Refresh timestamp on entry since it is used and return next hop address
        rt_update_entry(i);
        return routing_table[i].next_hop_address;
    } else if(aodv_check_discovery_in_progress(*dst) == 0) {
        //No valid route to host exists, and no discovery is running. Starting route discovery 
        
        //Create new RREQ packet 
        if((rreq = new_aodv_control_packet(RREQ_TYPE)) == NULL) {
            //Packet allocation failed, return with error
            ROUTING_DEBUG("\nE: (aodv_get_next_hop) Space allocation for RREQ failed.\n");
            return 0;
        } 
        rreq->flags &= 0; 
        rreq->id = ++aodv_rreq_id; //Every time a RREQ is emmited (also resend) id is incremented (Page 15)
        rreq->hopcount = 0; 
        rreq->type = RREQ_TYPE; 
        
        //Register adresses 
        rreq->destination_address = *dst; 
        rreq->originator_address = aodv_get_address(); 
        
        //Increment sequence number 
        /* This prevents conflicts with previously established 
            reverse routes.*/
        aodv_sequence_no++; 
        
        //Set originator (my own) sequence number 
        /* This must be done after the sequence number is increased */
        rreq->osn = aodv_sequence_no; 
        
        //Set the destination sequence number field in the rreq package. 
        /* The destination sequence number field in the RREQ  
            message is the last known destination sequence number 
            for this destination and is copied from the 
            destination sequence number field in the routing  
            table. If no sequence number is nown, the unknown 
            sequence number flag MUST be set. */
        
        if(i != -1) {
            //Invalid entry exists - clear flag and store sequence no
            rreq->flags &= (~FLAG_UNKNOWN_SEQUENCE_NUMBER);
            rreq->dsn = rt_get_sequence_no(i); //
                                               //IMPROVE - Set ttl for expanding ring search (and add code for expanding ring search of course)
            rreq->ttl = 0xFFFF;
            //rreq->ttl = UMAX(rt_get_hopcount(i) + TTL_INCREMENT, NET_DIAMETER); 
        } else {
            //Entry does not exist, set unknown sequence no flag
            rreq->flags |= FLAG_UNKNOWN_SEQUENCE_NUMBER;
            //IMPROVE - Set ttl for expanding ring search (and add code for expanding ring search of course)
            rreq->ttl = 0xFFFF; 
            //rreq->ttl = UMAX(TTL_START, NET_DIAMETER);
        }      
        
        //Add the route request to the discovery_buffer
        if(rreq_discovery_buffer_add(rreq) == -1) {
            //Could not add rreq to discovery buffer
            if(aodv_message_handler != NULL)
                aodv_message_handler(*dst, DISCOVERY_FAILED);
            return 0;
        }
        
        //Message upper layer that the route discovery will be started
        //(if a message handler was installed)
        if(aodv_message_handler != NULL)
            aodv_message_handler(*dst, DISCOVERY_STARTED);
        
        //Start the discovery
        NutEventPost(&event_rreq_pending);
    } //end new discovery
    
    //Next hop not yet known
    return 0; 
}



/**
* Returns the next hop on route if available,
 * otherwise starts route discovery if there is
 * not already one in progress.
 *
 * @param dst [in] destination address
 * @return next_hop_address on success, 0 otherwise
 *
 */
u_short aodv_get_next_hop(u_short *dst)
{
    u_short i; 
    
    // Look up destination in Table
    i = rt_get_entry(*dst);
    
    if(i != -1 && rt_check_entry_validity(i) && rt_check_entry_activity(i)) {
        //Refresh timestamp on entry since it is used and return next hop address
        rt_update_entry(i);
        return routing_table[i].next_hop_address;
    } else {
        //Next hop unknown
        return 0;
    } 
}


/**
* Makes route discovery
 *
 * @param dst [in] destination address
 * @return next_hop_address on success, 0 otherwise
 *
 */
THREAD(aodv_discover_thread,arg)
{
    int entry, retry_count=0;
    long t_wait;
    
    for(;;) {
    	 //printf("THREAD aodv_discover_thread\n");
    	 
        // Process next entry in queue if there is any.
        if(aodv_discovery_buffer_count > 0) {
            long t_next_shd = -1; // used to calculate next schedule for each packet.
            long curtime = NutGetMillis();
            
            
            //Aquire Mutex: We're going to work excessively on that buffer.
            NutMutexLock(&mutex_discovery_buffer);
            
            for(entry = 0; aodv_discovery_buffer_count > 0 && entry < DISCOVERY_BUFFER_SIZE; entry++) {
                
                if(aodv_discovery_buffer[entry].rreq != NULL) {
                    
                    /* is this packet ready to be resent? */
                    if (curtime >= aodv_discovery_buffer[entry].nextshd) {
                        //Entry ready, process
                        
                        if(aodv_use_led)
                            btn_led_set(LED0);
                        
                        //Buffer RREQ id and originator address for PATH_DISCOVERY_TIME 
                        /* In this way, when the node receives the packet 
                        again from its neighbours, it will not reprocess and  
                        reforward the packet. */
                        if(rreq_buffer_add(aodv_discovery_buffer[entry].rreq->id, 
                                           aodv_discovery_buffer[entry].rreq->originator_address) != 0) {
                            //Error, could not get space in Buffer. Abort discovery for this rreq;
                            ROUTING_DEBUG("\nE: (aodv_discover_thread) Failed buffering RREQ ID, aborting discovery\n"); 
                            
                            if(aodv_use_led)
                                btn_led_set(LED1);
                            
                            // Set error status
                            if(aodv_message_handler != NULL) 
                                aodv_message_handler(aodv_discovery_buffer[entry].rreq->destination_address,DISCOVERY_FAILED);
                            
                            // Drop packet
                            free(aodv_discovery_buffer[entry].rreq);
                            aodv_discovery_buffer[entry].rreq = NULL;
                            NutMutexLock(&mutex_discovery_buffer_count);
                            aodv_discovery_buffer_count--;
                            NutMutexUnlock(&mutex_discovery_buffer_count);
                            
                            if(aodv_use_led)
                                btn_led_clear(LED1);
                        } else {
                            //Send Packet out 
                            if(aodv_send_control_rreq_packet(BROADCAST_ADDR,aodv_discovery_buffer[entry].rreq) != 0) {
                                if(aodv_use_led) {
                                    btn_led_set(LED1);
                                    //IMPROVE - Add some delay to make LED flash visible
                                    btn_led_clear(LED1);
                                }
                            } else {
                                last_broadcast = NutGetMillis(); //Register last broadcast to decide if HELLO
                            }
                            
                            //Increase retry_count
                            retry_count = ++(aodv_discovery_buffer[entry].retry);
                            
                            //Decide if packet has reached its end of life
                            if(aodv_discovery_buffer[entry].retry == RREQ_RETRIES) {
                                ROUTING_DEBUG("\nW: (aodv_discover_thread) Max retry of rreq at pos %u reached.\n",entry);
                                
                                //Notify upper layer
                                if(aodv_message_handler != NULL) 
                                    aodv_message_handler(aodv_discovery_buffer[entry].rreq->destination_address,DISCOVERY_FAILED);
                                
                                //Remove the packet.
                                free(aodv_discovery_buffer[entry].rreq);
                                aodv_discovery_buffer[entry].rreq = NULL;
                                NutMutexLock(&mutex_discovery_buffer_count);
                                aodv_discovery_buffer_count--;
                                NutMutexUnlock(&mutex_discovery_buffer_count);
                            } else {
                                //Packet might be resent, update data
                                aodv_discovery_buffer[entry].rreq->id = ++aodv_rreq_id;
                                //I have no clue if we must adjust sequence number (RFC unclear). I'd say no.
                                
                                //Calculate next schedule for this packet
                                //IMPROVE - Add code for expanding ring search here.
                                //aodv_discovery_buffer[entry].nextshd = NutGetMillis() + aodv_beb(NET_TRAVERSAL_TIME, retry_count);
                                
                                // no binary exponential backoff for this
                                aodv_discovery_buffer[entry].nextshd = NutGetMillis() + (rand() % NET_TRAVERSAL_TIME) + NET_TRAVERSAL_TIME;
                            }
                        }
                        
                        if(aodv_use_led)
                            btn_led_clear(LED0);
                    } //end send scheduled packet 
                    
                    //Packet is valid, but not yet ready to be sent (not scheduled)
                    //so we calculate the next sleep-time. do this by getting the
                    //timeout minimum of all packets in the queue 
                    if (aodv_discovery_buffer[entry].rreq != NULL && //Packet might have been deleted
                        (aodv_discovery_buffer[entry].nextshd < t_next_shd || t_next_shd < curtime))
                        t_next_shd = aodv_discovery_buffer[entry].nextshd;
                } // end if valid entry
            } // end for
            
            //Unlock Mutex
            NutMutexUnlock(&mutex_discovery_buffer);
            
            //Retry to send packets after timeout or when new packet is put in queue
            //make sure, that we don't sleep forever
            t_wait = t_next_shd - NutGetMillis();
            if (t_wait > 0) { 
                NutEventWait(&event_rreq_pending, t_wait);
            }
        } else {
            // Wait until there is something to send
            NutEventWait(&event_rreq_pending, NUT_WAIT_INFINITE);
        } 
    }
}


/**
* Checks if a route discovery is in progress
 *
 * @param [in] destination for which a discovery is searched
 * @return 1 if a discovery is in progress, 0 otherwise
 */
static inline u_char aodv_check_discovery_in_progress(u_short dst)
{
    int j;
    
    //Check if there is a route discovery in progress for this destination
    NutMutexLock(&mutex_discovery_buffer);
    for(j = 0; j < DISCOVERY_BUFFER_SIZE; j++) {
        if(aodv_discovery_buffer[j].rreq != NULL && 
           aodv_discovery_buffer[j].rreq->destination_address == dst)
            break;
    }
    NutMutexUnlock(&mutex_discovery_buffer);
    
    if(j == DISCOVERY_BUFFER_SIZE) 
        return 0;
    else
        return 1;
} 


//-------------------------------
//Sending Functionallity
//-------------------------------

/** 
* Sends a RREQ package to a given destination
*
* @param addr [in] address where the packet should be sent to
* @param rreq [in] pointer to the packet
* @return 0 on success
*/
static int aodv_send_control_rreq_packet(u_short addr, aodv_control_rreq_packet_t *rreq)
{ 
    ccc_packet_t *cpkt; 
    
    aodv_rreq_send_count++;
    
    //Create new ccc packet 
    if(( (cpkt = new_ccc_packet(sizeof(aodv_control_rreq_packet_t))) == 0) ) {
        //Allocation Failed
        ROUTING_DEBUG("\nE: (aodv_send_control_rreq_packet) Space allocation failed!\n");
        aodv_rreq_send_fail_count++;
        return -1;
    }
    
    //Enter ccc header information 
    cpkt->length = sizeof(aodv_control_rreq_packet_t);
    
    //Serialise payload
    aodv_pkt_serialize_rreq(cpkt->data, rreq);
    
    //Send RREQ to addr
    if(ccc_send(addr, CONTROL_TYPE, cpkt)) { 
        //send failed 
        aodv_rreq_send_fail_count++;
        free(cpkt);
        return -1; 
    } else { 
        ROUTING_DEBUG("\n>>> RREQ to %u", addr);
        free(cpkt);
        return 0; 
    } 
    
} 

/** 
* Sends a RREP package to a given destination
*
* @param addr [in] address where the packet should be sent to
* @param rrep [in] pointer to the packet
* @return 0 on success
*/
static int aodv_send_control_rrep_packet(u_short addr, aodv_control_rrep_packet_t *rrep)
{ 
    ccc_packet_t *cpkt; 
    
    aodv_rrep_send_count++;
    
    //Create new ccc packet 
    if(( (cpkt = new_ccc_packet(sizeof(aodv_control_rreq_packet_t))) == 0) ) {
        //Allocation Failed
        ROUTING_DEBUG("\nE: (aodv_send_control_rrep_packet) Space allocation failed!\n");
        aodv_rrep_send_fail_count++;
        return -1;
    }
    
    //Enter ccc header information 
    cpkt->length = sizeof(aodv_control_rrep_packet_t);
    
    //Serialise payload
    aodv_pkt_serialize_rrep(cpkt->data, rrep);
    
    //Send RREQ to addr
    if(ccc_send(addr, CONTROL_TYPE, cpkt)) { 
        //send failed 
        aodv_rrep_send_fail_count++;
        free(cpkt);
        return -1; 
    } else { 
        ROUTING_DEBUG("\n>>> RREP to %u",addr);
        free(cpkt);
        return 0; 
    } 
    
} 


/** 
* Sends a RREP ACK package to a given destination
*
* @param addr [in] address where the packet should be sent to
* @param rrep_ack [in] pointer to the packet 
* @return 0 on success
*/

static int aodv_send_control_rrep_ack_packet(u_short addr, aodv_control_rrep_ack_packet_t *rrep_ack)
{ 
    ccc_packet_t *cpkt; 
    
    aodv_rrep_ack_send_count++;
    
    //Create new ccc packet 
    if(( (cpkt = new_ccc_packet(sizeof(aodv_control_rrep_ack_packet_t))) == 0) ) {
        //Allocation Failed
        ROUTING_DEBUG("\nE: (aodv_send_control_rrep_ack_packet) Space allocation failed!\n");
        aodv_rrep_ack_send_fail_count++;
        return -1;
    }
    
    //Enter ccc header information 
    cpkt->length = sizeof(aodv_control_rrep_ack_packet_t);
    
    //Serialise payload
    aodv_pkt_serialize_rrep_ack(cpkt->data, rrep_ack);
    
    //Send RREQ to addr
    if(ccc_send(addr, CONTROL_TYPE, cpkt)) { 
        //send failed 
        aodv_rrep_ack_send_fail_count++;
        free(cpkt);
        return -1; 
    } else { 
        ROUTING_DEBUG("\n>>> RREP-ACK to %u",addr);
        free(cpkt);
        return 0; 
    } 
    
} 

/** 
* Sends a RRER package to a given destination
*
* @param addr [in] address where packet should be sent to
* @param rerr [in] pointer to the rerr packet that should be sent
* @return 0 on success
*/
static int aodv_send_control_rerr_packet(u_short addr, aodv_control_rerr_packet_t *rerr)
{ 
    ccc_packet_t *cpkt; 
    
    aodv_rerr_send_count++;
    
    //Create new ccc packet 
    //IMPROVE - Make allocation dinamycally to support more than two destinations.
    if(( (cpkt = new_ccc_packet(sizeof(aodv_control_rerr_packet_t))) == 0) ) {
        //Allocation Failed
        ROUTING_DEBUG("\nE: (aodv_send_control_rerr_packet) Space allocation failed!\n");
        aodv_rerr_send_fail_count++;
        return -1;
    }
    
    //Enter ccc header information 
    cpkt->length = sizeof(aodv_control_rerr_packet_t);
    
    //Serialise payload
    aodv_pkt_serialize_rerr(cpkt->data, rerr);
    
    //Send RERR to addr
    if(ccc_send(addr, CONTROL_TYPE, cpkt)) { 
        //send failed 
        aodv_rerr_send_fail_count++;
        free(cpkt);
        return -1; 
    } else { 
        ROUTING_DEBUG("\n>>> RERR to %u",addr);
        free(cpkt);
        return 0; 
    } 
    
} 





/**
* Sends RREPs until ACK received or MAX_RREQ_RETRIES reached
 */
THREAD(aodv_rrep_thread,arg)
{
    int entry, retry_count = 0;
    long t_wait = 0;
    
    for(;;) {
    	
    	//printf("THREAD aodv_rrep_thread\n");
        
        // Process next entry in queue if there is any.
        NutMutexLock(&mutex_rrep_buffer_count);
        
        if(aodv_rrep_buffer_count > 0) {
            long t_next_shd = -1; //Used to calculate next schedule for each packet
            long curtime = NutGetMillis();
            
            NutMutexUnlock(&mutex_rrep_buffer_count);
            
            //Aquire Mutex: We're going to work excessively on that buffer.
            NutMutexLock(&mutex_rrep_buffer);
            
            for(entry = 0; aodv_rrep_buffer_count > 0 && entry < RREP_BUFFER_SIZE; entry++) {
                
                if(aodv_rrep_buffer[entry].rrep != NULL) {
                    
                    //Packet ready?
                    if (curtime>=aodv_rrep_buffer[entry].nextshd) {
                        
                        //Entry ready, resend rrep	    
                        aodv_send_control_rrep_packet(aodv_rrep_buffer[entry].dst, aodv_rrep_buffer[entry].rrep);
                        retry_count = ++(aodv_rrep_buffer[entry].retry);
                        
                        //Check if max retry reached
                        if(aodv_rrep_buffer[entry].retry == RREP_RETRY) {
                            free(aodv_rrep_buffer[entry].rrep);
                            aodv_rrep_buffer[entry].rrep = NULL;
                            NutMutexLock(&mutex_rrep_buffer_count);
                            aodv_rrep_buffer_count--;
                            NutMutexUnlock(&mutex_rrep_buffer_count);
                            
                            //Add target node to blacklist
                            aodv_blacklist_add(aodv_rrep_buffer[entry].dst);
                        } else {
                            //Calculate next schedule
                            aodv_rrep_buffer[entry].nextshd = NutGetMillis() + aodv_beb(NEXT_HOP_WAIT, retry_count); 
                        }
                        
                    } //end send scheduled packet 
                    
                    //Packet is valid, but not yet ready to be sent (not scheduled)
                    //so we calculate the next sleep-time. we do this by getting the
                    //timeout minimum of all packets in the queue 
                    if (aodv_rrep_buffer[entry].rrep != NULL && //Packet might have been deleted
                        (aodv_rrep_buffer[entry].nextshd < t_next_shd || t_next_shd < curtime))
                        t_next_shd = aodv_rrep_buffer[entry].nextshd;
                } // end if valid entry
            } // end for
            
            //Unlock Mutex
            NutMutexUnlock(&mutex_rrep_buffer);
            
            //Retry to send packets after timeout or when new packet is put in queue
            //make sure, that we don't sleep forever
            t_wait = t_next_shd - NutGetMillis();
            if (t_wait > 0) {
                NutEventWait(&event_rrep_pending, t_wait);
            }
        } else {
            NutMutexUnlock(&mutex_rrep_buffer_count);
            
            // Wait until there is something to send
            NutEventWait(&event_rrep_pending, NUT_WAIT_INFINITE);
        } 
    }
}



//-------------------------------
//Receiving Functionality
//-------------------------------

/**
* Checks the control queue for new entries and processes them
 *
 */
THREAD(recv_control_thread, arg)
{
    u_short                    pos = 0;      //Points to the next packet which should be processed
    aodv_control_rrep_packet_t *rrep = NULL;
    aodv_control_rreq_packet_t *rreq = NULL;

	u_short packet_count = 0;
    
    u_short i = 0, j = 0;
    
    
    for(;;) {
    	
    // quick 'n dirty: process all packets
    
    NutEventWait(&event_control_packet_recvd, NUT_WAIT_INFINITE);
    //printf("THREAD recv_control_thread (counter=%u)\n", aodv_recv_control_packet_buffer_counter);
	packet_count = 0;
    for ( pos = 0; pos < CONTROL_PACKET_QUEUE_SIZE; pos++ )
	{    
    	
	    	
        
#ifdef ROUTING_DEBUGMODE
        //Preconditions
        if(rreq != NULL) {
            ROUTING_DEBUG("\nW: (recv_control_thread) Memory leak possible, RREQ != NULL\n");
        }
        if(rrep != NULL) {
            ROUTING_DEBUG("\nW: (recv_control_thread) Memory leak possible, RREP != NULL\n");
        }
#endif //ROUTING_DEBUGMODE
        
        /*
        //Check if there is an entry in the control packet receive queue (only reading, thus no mutex)
        if(aodv_recv_control_packet_buffer_counter == 0) {
            //No packets waiting, sleep and wait for signal
            NutEventWait(&event_control_packet_recvd, NUT_WAIT_INFINITE);
        } else 
        
        {	
        */
        
            //Packets waiting, process them
            
            //Check if fetch points to a valid packet
            if(aodv_recv_control_packet_buffer[pos].pkt != NULL) {
                packet_count++;
                
                switch(aodv_recv_control_packet_buffer[pos].pkt->data[0]) {
                    
                    case RREQ_TYPE:
                        /*****************************************************************/
                        /*                               RREQ                            */
                        /*****************************************************************/	  
                        
                        ROUTING_DEBUG("\n<<< RREQ from %u",aodv_recv_control_packet_buffer[pos].src);
                        aodv_rreq_recv_count++;
                        
                        //If leds are on, flash green
                        if(aodv_use_led)
                            btn_led_set(LED0);
                            
                            //Alocate new aodv rreq packet, if the node is not blacklisted. Otherwise ignore.
                            if( aodv_blacklist_lookup(aodv_recv_control_packet_buffer[pos].src) ||
                                (rreq = new_aodv_control_packet(RREQ_TYPE)) == NULL) {
                                
                                if(aodv_use_led)
                                    btn_led_set(LED1);
                                
                                aodv_rreq_recv_fail_count++;
                                
#ifdef ROUTING_INFOMODE
                                if(aodv_blacklist_lookup(aodv_recv_control_packet_buffer[pos].src)) {
                                    //Node is blacklisted
                                    ROUTING_DEBUG("\nI: (recv_control_thread) Node %u is blacklisted.\n", 
                                                  aodv_recv_control_packet_buffer[pos].src);
                                }
#endif //ROUTING_INFOMODE
                                
#ifdef ROUTING_DEBUGMODE
                                if(aodv_blacklist_lookup(aodv_recv_control_packet_buffer[pos].src) == 0 && rreq == NULL) {
                                    //Space allocation failed, skip
                                    ROUTING_DEBUG("\nE: (recv_control_thread) Space allocation for receiving rreq failed!\n");
                                }
#endif //ROUTING_DEBUGMODE
                                
                                if(aodv_use_led)
                                    btn_led_clear(LED1);
                                
                            } else {
                                //Deserialize Payload
                                aodv_pkt_deserialize_rreq(rreq, aodv_recv_control_packet_buffer[pos].pkt->data);
                                
                                //Check if an entry to the source exists in the routing table. If not one
                                //is created:
                                /* When a node receives a RREQ, it first creates or updates a route to the previous 
                                hop without a valid sequence number (6.5) */
                                if((i = rt_get_entry(aodv_recv_control_packet_buffer[pos].src)) != -1) {
                                    //entry exists, update it
                                    NutMutexLock(&mutex_routing_table);
                                    routing_table[i].lifetime = NutGetMillis() + ACTIVE_ROUTE_TIMEOUT;
                                    NutMutexUnlock(&mutex_routing_table);
                                    
                                } else {
                                    //Entry does not exist, create a new one
                                    rt_create_entry(aodv_recv_control_packet_buffer[pos].src,
                                                    0,
                                                    0,  
                                                    1, 
                                                    aodv_recv_control_packet_buffer[pos].src, 
                                                    ACTIVE_ROUTE_TIMEOUT);
                                }
                                
                                
                                //Check if this packet was already seen. If so, drop it. Otherwise add it to the queue
                                //and if successfully added to the queue, process it
                                /* In this way, when the node receives the packet 
                                    again from its neighbours, it will not reprocess and  
                                    reforward the packet. */
                                if(rreq_buffer_lookup(rreq->id, rreq->originator_address) != 0 &&    //Not yet seen
                                   rreq_buffer_add(rreq->id, rreq->originator_address) == 0) {       //Added successfully
                                                                                                     //Process Packet
                                    
                                    //Increment the hop count
                                    /* First, it first increments the hop count value in the RREQ by one... (6.5) */
                                    rreq->hopcount++;
                                    
                                    //Search for Reverse route to originator
                                    /* Then the node searches for a reverse route to the originator IP address (see section
                                    6.2), using longest-prefix matching. If need be, the route is created, or updated using
the originator sequence number from the rreq in its routing table. (6.5) */
                                    i = rt_get_entry(rreq->originator_address);
                                    if((signed)i != -1) {
                                        //Reverse rexists, update it		
                                        NutMutexLock(&mutex_routing_table);
                                        
                                        /* the Originator Sequence Number from the RREQ is compared to the corresponding destination
                                        sequence number in the route table entry and copied if greater thant the existing value there */
                                        if(rreq->osn > routing_table[i].dsn)
                                            routing_table[i].dsn = rreq->osn;
                                        
                                        /* the valid sequence number field is set to true */
                                        routing_table[i].flags |= FLAG_VALID; 
                                        
                                        /* the next hop in the routing table becomes the node from which the RREQ was received ... */
                                        routing_table[i].next_hop_address = aodv_recv_control_packet_buffer[pos].src;
                                        
                                        /* 4. the hop count is copied from the Hop Count in the RREQ message; */
                                        routing_table[i].metric = rreq->hopcount;
                                        
                                        MAX( routing_table[i].lifetime, 
                                             NutGetMillis() + 2 * NET_TRAVERSAL_TIME - 2 * rreq->hopcount * NODE_TRAVERSAL_TIME);
                                        
                                        NutMutexUnlock(&mutex_routing_table);
                                        
                                    } else {
                                        //Entry does not exist, establish reverse route
                                        // --> Note: Below is an extract from the RFC, however, if we do not assume symmetric routes,
                                        //           we really should initiate a new route rrequest here, since only assuming the 
                                        //           previous hop to be reachable is not always true.
                                        rt_create_entry(rreq->originator_address,
                                                        rreq->osn,
                                                        FLAG_VALID | FLAG_ACTIVE,
                                                        rreq->hopcount,
                                                        aodv_recv_control_packet_buffer[pos].src, 
                                                        ACTIVE_ROUTE_TIMEOUT);
                                    }
                                    
                                    // Decide what action to take:
                                    //  1. Reply if this is the final destination
                                    //  2. Reply if this node knows a route to the final destination
                                    //  3. Update and Rebroadcast the RREQ
                                    if(rreq->destination_address == aodv_get_address()) {
                                        //Case 1: I am the final destination, so send a RREP
                                        if( (rrep = new_aodv_control_packet(RREP_TYPE)) == NULL) {
                                            //Space allocation failed
                                            ROUTING_DEBUG("\nE: (recv_control_thread) Could not allocate space for RREP!\n");
                                        } else {
                                            //Increment Sequence number
                                            /* If the generating node is the destination itself, it MUST increment
                                            its own sequence number by one if the sequence number in the RREQ
                                            packet is equal to that incremented value. (6.6.1) */
                                            if(rreq->dsn == aodv_sequence_no + 1)
                                                aodv_sequence_no++;
                                            
                                            //Fill in Data
                                            rrep->type = RREP_TYPE;
                                            if(use_ack) 
                                                rrep->flags = FLAG_ACK;  // it must be = to initialize flags atm
                                            else 
                                                rrep->flags = 0;
                                            rrep->hopcount = 0;
                                            rrep->destination_address = aodv_address; 
                                            rrep->dsn = aodv_sequence_no; 
                                            rrep->originator_address = rreq->originator_address;
                                            rrep->lifetime = (u_long)MY_ROUTE_TIMEOUT; 
                                            rrep->ttl = 0xFFFF; //IMPROVE - Sett ttl acording to information in RREQ
                                            
                                            
                                            //Send
                                            if(use_ack) {
                                                //ACKS needed, add to buffer
                                                if( rrep_buffer_add(rrep, aodv_recv_control_packet_buffer[pos].src) == -1 ) {
                                                    //If leds are used, flash green
                                                    if(aodv_use_led)
                                                        btn_led_set(LED3);
                                                    
                                                    ROUTING_DEBUG("\nW: (recv_control_thread) Could not add rrep to buffer, sending unbuffered");
                                                    rrep->flags &= (~FLAG_ACK); 
                                                    aodv_send_control_rrep_packet(aodv_recv_control_packet_buffer[pos].src, rrep);
                                                    
                                                    //Clear led
                                                    if(aodv_use_led)
                                                        btn_led_clear(LED3);
                                                } else {
                                                    NutEventPost(&event_rrep_pending);
                                                }
                                            } else  {
                                                //No need for ACK, send unbuffered
                                                
                                                //If leds are used, flash green
                                                if(aodv_use_led)
                                                    btn_led_set(LED3);
                                                
                                                aodv_send_control_rrep_packet(aodv_recv_control_packet_buffer[pos].src, rrep);
                                                
                                                //Clear led
                                                if(aodv_use_led)
                                                    btn_led_clear(LED3);
                                            }
                                            //If sequence number was not incremented before send, do this now
                                            if(rreq->dsn != aodv_sequence_no) 
                                                aodv_sequence_no++;
                                            
                                            //Free space
#ifdef ROUTING_DEBUGMODE
                                            if(rrep == NULL)
                                                ROUTING_DEBUG("\nA: (recv_control_thread) Freeing rrep that is NULL!\n");
#endif //ROUTING_DEBUGMODE
                                            free(rrep);
                                            rrep = NULL;
                                        }
                                        
                                    } else {
                                        //Case 2 or 3
                                        i = rt_get_entry(rreq->destination_address);
                                        
                                        if((signed)i != -1 && 
                                           rt_check_entry_activity(i) && 
                                           rt_check_entry_validity(i) && 
                                           (rreq->flags & FLAG_UNKNOWN_SEQUENCE_NUMBER || routing_table[i].dsn >= rreq->dsn)) {
                                            //Case 2: I know a valid, fresh route to the final destination, send RREP
                                            
                                            if( (rrep = new_aodv_control_packet(RREP_TYPE)) == NULL) {
                                                //Space allocation failed
                                                ROUTING_DEBUG("\nE: (recv_control_thread) Could not allocate space for RREP!\n");
                                            } else {
                                                //IMPROVE - If you plan to use precursor lists, update them here for the forward route entry
                                                
                                                //Fill in Data
                                                rrep->type = RREP_TYPE;
                                                if(use_ack) 
                                                    rrep->flags = FLAG_ACK;
                                                else
                                                    rrep->flags = 0; 
                                                rrep->hopcount = routing_table[i].metric;
                                                rrep->destination_address = routing_table[i].dst_address; 
                                                rrep->dsn = routing_table[i].dsn; 
                                                rrep->originator_address = rreq->originator_address;
                                                rrep->lifetime = routing_table[i].lifetime - NutGetMillis(); 
                                                rrep->ttl = 0xFFFF; 
                                                
                                                //Update Precursorlist in reverse route entry
                                                j = rt_get_entry(rreq->originator_address);
                                                
                                                //IMPROVE - If you plan to use precursor lists, update them here
                                                
                                                
                                                //Send that RREP
                                                if(j != -1 && rt_check_entry_validity(j) && rt_check_entry_activity(j)) {
                                                    //Send packet to next hop on forward route to originator of RREQ
                                                    
                                                    if(use_ack) {
                                                        //ACKS needed, add to buffer
                                                        if( rrep_buffer_add(rrep, aodv_recv_control_packet_buffer[pos].src) == -1 ) {
                                                            //If leds are used, flash green
                                                            if(aodv_use_led)
                                                                btn_led_set(LED3);
                                                            
                                                            ROUTING_DEBUG("\nW: (recv_control_thread) Could not add rrep to buffer, sending unbuffered");
                                                            rrep->flags &= (~FLAG_ACK); 
                                                            aodv_send_control_rrep_packet(routing_table[j].next_hop_address, rrep);
                                                            
                                                            //Clear led
                                                            if(aodv_use_led)
                                                                btn_led_clear(LED3);
                                                        } else {
                                                            NutEventPost(&event_rrep_pending);
                                                        }
                                                    } else {
                                                        
                                                        //If leds are used, flash green
                                                        if(aodv_use_led)
                                                            btn_led_set(LED3);
                                                        
                                                        aodv_send_control_rrep_packet(routing_table[j].next_hop_address, rrep);
                                                        
                                                        //Clear led
                                                        if(aodv_use_led)
                                                            btn_led_clear(LED3);
                                                    }
                                                } else { 
                                                    //Send packet to node that sent the rreq
                                                    
                                                    if(use_ack) {
                                                        //ACKS needed, add to buffer
                                                        if( rrep_buffer_add(rrep, aodv_recv_control_packet_buffer[pos].src) == -1 ) {
                                                            //If leds are used, flash green
                                                            if(aodv_use_led)
                                                                btn_led_set(LED3);
                                                            
                                                            ROUTING_DEBUG("\nW: (recv_control_thread) Could not add rrep to buffer, sending unbuffered!\n");
                                                            rrep->flags &= (~FLAG_ACK); 
                                                            aodv_send_control_rrep_packet(aodv_recv_control_packet_buffer[pos].src, rrep);
                                                            
                                                            //Clear led
                                                            if(aodv_use_led)
                                                                btn_led_clear(LED3);
                                                        } else {
                                                            NutEventPost(&event_rrep_pending);
                                                        }
                                                    } else {
                                                        //Send unbuffered
                                                        
                                                        //If leds are used, flash green
                                                        if(aodv_use_led)
                                                            btn_led_set(LED3);
                                                        
                                                        aodv_send_control_rrep_packet(aodv_recv_control_packet_buffer[pos].src, rrep); 
                                                        
                                                        //Clear led
                                                        if(aodv_use_led)
                                                            btn_led_clear(LED3);
                                                    }
                                                }
                                                
                                                //If "G" flag is set in the RREQ, also send a gratious RREP to the destination
                                                if(rreq->flags & FLAG_GRATIOUS) {
                                                    //Fill in Data
                                                    rrep->type = RREP_TYPE;
                                                    if(use_ack) rrep->flags = FLAG_ACK;
                                                    rrep->hopcount = routing_table[i].metric;
                                                    rrep->destination_address = rreq->originator_address; 
                                                    rrep->dsn = rreq->osn; 
                                                    rrep->originator_address = rreq->destination_address;
                                                    rrep->lifetime = routing_table[i].lifetime - NutGetMillis();
                                                    rrep->ttl = 0xFFFF;
                                                    
                                                    //Send that gratious RREP
                                                    //IMPROVE - enable ACK queueing for gratious RREPs... (as above)
                                                    aodv_send_control_rrep_packet(routing_table[i].next_hop_address, rrep);
                                                }
                                                
                                                //Free rrep space
#ifdef ROUTING_DEBUGMODE
                                                if(rrep == NULL)
                                                    ROUTING_DEBUG("\nA: (recv_control_thread) Freeing rrep that is already NULL.\n");
#endif //ROUTING_DEBUGMODE
                                                free(rrep);
                                                rrep = NULL;
                                            } //endif allocation failed 
                                            
                                        } else if(rreq->ttl > 1) { 
                                            //I do not know a route, however, the TTL of the RREQ is bigger than 1
                                            //I update and broadcast the RREQ
                                            rreq->ttl--; 
                                            
                                            if( (rreq->flags & FLAG_UNKNOWN_SEQUENCE_NUMBER) == 0 && 
                                                (signed)i != -1 && (routing_table[i].flags & FLAG_VALID) ) 
                                                rreq->dsn = UMAX((short)routing_table[i].dsn,(short)rreq->dsn);
                                            else if( i != -1 && (routing_table[i].flags & FLAG_VALID) ) {
                                                rreq->dsn = routing_table[i].dsn;
                                                rreq->flags &= (~FLAG_UNKNOWN_SEQUENCE_NUMBER);
                                            }
                                            
                                            //Broadcast the RREQ
                                            if( (i = aodv_send_control_rreq_packet(BROADCAST_ADDR,rreq)) == 0) {
                                                //Broadcast successfull
                                                last_broadcast = NutGetMillis();
                                            }
                                            
                                        }
                                    }
                                } 
                                else {
#ifdef ROUTING_INFOMODE
                                    ROUTING_DEBUG("\nI: (recv_control_thread) RREQ already seen. *CRUNCH*\n");
#endif // ROUTING_INFOMODE
                                    aodv_rreq_ignore_count++;
                                }
                                
                                //Free the route request
#ifdef ROUTING_DEBUGMODE
                                if(rreq == NULL)
                                    ROUTING_DEBUG("\nA: (recv_control_thread) Going to free rreq that is already NULL.\n");
#endif //ROUTING_DEBUGMODE
                                free(rreq);
                                rreq = NULL;
                            }
                                
                                //Clear led
                                if(aodv_use_led)
                                    btn_led_clear(LED0);
                                
                                break;
                        
                        
                    case RREP_TYPE:
                        /*****************************************************************/
                        /*                               RREP                            */
                        /*****************************************************************/	  
                        
                        ROUTING_DEBUG("\n<<< RREP from %u",aodv_recv_control_packet_buffer[pos].src);
                        aodv_rrep_recv_count++
                            ;
                        //If leds are on, flash green
                        if(aodv_use_led)
                            btn_led_set(LED3);
                            
                            
                            //Alocate new aodv rrep packet
                            if( (rrep = new_aodv_control_packet(RREP_TYPE)) == NULL) {
                                //Space allocation failed, skip
                                ROUTING_DEBUG("\nE: (recv_control_thread) Could not allocate space for receiving rrep!\n");
                                aodv_rrep_recv_fail_count++;
                            } else {
                                //Deserialize Payload 
                                aodv_pkt_deserialize_rrep(rrep, aodv_recv_control_packet_buffer[pos].pkt->data);
                                
                                //If an ACK is requeseted, send one (we should do this right now, no more delays)
                                if(rrep->flags & FLAG_ACK) {
                                    //Send ACK
                                    aodv_control_rrep_ack_packet_t *rrep_ack = new_aodv_control_packet(RREP_ACK_TYPE);
                                    if(rrep_ack != NULL) {
                                        rrep_ack->type = RREP_ACK_TYPE;
                                        aodv_send_control_rrep_ack_packet(aodv_recv_control_packet_buffer[pos].src,rrep_ack);
                                        free(rrep_ack);
                                        rrep_ack = NULL;
                                    } else {
                                        ROUTING_DEBUG("\nE: (recv_control_thread) Space allocation for rrep ack failed. Not sending.\n");
                                    }
                                }
                                
                                //Check if an entry to the source exists in the routing table. If not one
                                //is created:
                                /* When a node receives a RREP message, it searches (using longest prefix matching
                                    for a route to the previous hop. If needed, a route is created for the previous
                                    hop, but without a valid sequence number (see section 6.2). (6.7) */
                                if((signed)(i = rt_get_entry(aodv_recv_control_packet_buffer[pos].src)) == -1) {
                                    //Entry does not exist, create a new one
                                    rt_create_entry(aodv_recv_control_packet_buffer[pos].src,
                                                    0,
                                                    0, 
                                                    1, 
                                                    aodv_recv_control_packet_buffer[pos].src,
                                                    ACTIVE_ROUTE_TIMEOUT);
                                }
                                
                                //Increment the hopcount
                                /* Next the node increments the hop count value in the rrep by one */
                                rrep->hopcount++;
                                
                                //Create forward route to destination, if it does not already exist. If it does
                                //compare the DSN in the mesage with the dsn in the rt entry.
                                /* Then the forward route for this destination is created if it does not already exist.
                                    Otherwise, the node compares the Destination Sequence Number in the message with its
                                    own stored destination sequence number for the Destinations IP Address in the RREP
                                    message. (6.7)*/
                                j = 0; //Shows wheter an update occured or not
                                if((signed)(i = rt_get_entry(rrep->destination_address)) == -1) {
                                    j = 1;
                                    //Entry does not exist, create one
                                    rt_create_entry(rrep->destination_address,
                                                    rrep->dsn,
                                                    FLAG_ACTIVE | FLAG_VALID, 
                                                    rrep->hopcount, 
                                                    aodv_recv_control_packet_buffer[pos].src,
                                                    NutGetMillis() + rrep->lifetime);
                                    
                                } else if ( ( rt_check_entry_validity(i) == 0 ) ||
                                            ( rrep->dsn > routing_table[i].dsn && rt_check_entry_validity(i) ) ||
                                            ( rrep->dsn == routing_table[i].dsn && (rt_check_entry_activity(i)==0) ) ||
                                            ( rrep->dsn == routing_table[i].dsn && rrep->hopcount < routing_table[i].metric) ) {
                                    j = 1;
                                    
                                    //Entry exists and must be updated, update it
                                    NutMutexLock(&mutex_routing_table);
                                    routing_table[i].flags |= FLAG_VALID | FLAG_ACTIVE;
                                    routing_table[i].next_hop_address = aodv_recv_control_packet_buffer[pos].src;
                                    routing_table[i].metric = rrep->hopcount;
                                    routing_table[i].lifetime = NutGetMillis() + rrep->lifetime;
                                    routing_table[i].dsn = rrep->dsn;
                                    NutMutexUnlock(&mutex_routing_table);
                                }
                                
                                //If a route was updated and I am not the originator, forward RREP
                                if(j == 1 && aodv_address != rrep->originator_address && rrep->ttl > 1) {
                                    //Decrease TTL
                                    rrep->ttl--;
                                    
                                    //Look for next hop towards originator
                                    if((signed)(i = rt_get_entry(rrep->originator_address)) == -1) {
                                        //Entry does not exist, report error
                                        ROUTING_DEBUG("\nW: (recv_control_thread) Could not find next hop to forward RREP.\n");
                                    } else {
                                        //Forward RREP towards destination
                                        if(use_ack) {
                                            //Set ack flag
                                            rrep->flags |= FLAG_ACK;
                                            
                                            //ACKS needed, add to buffer
                                            if( rrep_buffer_add(rrep, routing_table[i].next_hop_address) == -1 ) {
                                                ROUTING_DEBUG("\nW: (recv_control_thread) Could not add rrep to buffer, sending unbuffered.\n");
                                                rrep->flags &= (~FLAG_ACK); 
                                                aodv_send_control_rrep_packet(routing_table[i].next_hop_address, rrep);
                                            } else {
                                                NutEventPost(&event_rrep_pending);
                                            }
                                        } else  {
                                            //No need for ACK, send unbuffered
                                            aodv_send_control_rrep_packet(routing_table[i].next_hop_address, rrep);
                                        }
                                        
                                        //IMPROVE - If you decide to use precursor lists,  update them here
                                        /* When any node transmits a RREP, the precursor list for the corresponding destination
                                        node is updated by adding to it the next hop node to which the RREP is forwarded. (6.7)*/
                                        
                                        //Adjust lifetime of the reverse route that was used to forward the RREP
                                        NutMutexLock(&mutex_routing_table);
                                        routing_table[i].lifetime = UMAX(routing_table[i].lifetime, NutGetMillis() + ACTIVE_ROUTE_TIMEOUT);
                                        NutMutexUnlock(&mutex_routing_table);
                                        
                                        //IMPROVE- If you decide to use precursor lists, update them here (2)
                                        /* Finally, the precursor list for the next hop towards the destination is updated to contain the
                                            next hop towards the source. (6.7) */
                                    }
                                } else if (aodv_address == rrep->originator_address) {
                                    //I am the originator of the rreq, delete rreq from discovery buffer
                                    
                                    //Delete rreq that looks for destination in RREP
                                    NutMutexLock(&mutex_discovery_buffer);
                                    for(i = 0; i < DISCOVERY_BUFFER_SIZE; i++) {
                                        if(aodv_discovery_buffer[i].rreq != NULL &&
                                           aodv_discovery_buffer[i].rreq->destination_address == rrep->destination_address)
                                            break;
                                    }
                                    if(i != DISCOVERY_BUFFER_SIZE) {
                                        //Delete
                                        free(aodv_discovery_buffer[i].rreq);
                                        aodv_discovery_buffer[i].rreq = NULL;
                                        NutMutexLock(&mutex_discovery_buffer_count);
                                        aodv_discovery_buffer_count--;
                                        NutMutexUnlock(&mutex_discovery_buffer_count);
                                    }
#ifdef ROUTING_INFOMODE
                                    else {
                                        ROUTING_DEBUG("\nI: (recv_control_thread) Packet somehow disapeared from discovery buffer. That's ok.\n");
                                    }
#endif //ROUTING_INFOMODE
                                    NutMutexUnlock(&mutex_discovery_buffer);
                                    
                                    //Message upper layer that route discovery was successfull
                                    if(aodv_message_handler != NULL) 
                                        aodv_message_handler(rrep->destination_address,ROUTE_AVAILABLE);
                                }
#ifdef ROUTING_INFOMODE
                                else {
                                    if(rrep->ttl > 1)
                                        ROUTING_DEBUG("\nI: (recv_control_thread) RREP received but not processed (old info?)\n");
                                }
#endif //ROUTING_INFOMODE
                                
                                //Free the rrep 
#ifdef ROUTING_DEBUGMODE
                                if(rrep == NULL)
                                    ROUTING_DEBUG("\nA: (recv_control_thread) Going to free RREP that is already NULL.\n");
#endif //ROUTING_DEBUGMODE
                                free(rrep); 
                                rrep = NULL; 
	  } 
                                
                                //Clear led, if used
                                if(aodv_use_led)
                                    btn_led_clear(LED3);
                                
                                break;
                        
                    case RREP_ACK_TYPE:
                        /*****************************************************************/
                        /*                           RREP ACK                            */
                        /*****************************************************************/	  
                        
                        ROUTING_DEBUG("\n<<<RREP-ACK from %u",aodv_recv_control_packet_buffer[pos].src);
                        aodv_rrep_ack_recv_count++;
                        
                        //Delete Pending ACK from ACK queue
                        rrep_buffer_del(aodv_recv_control_packet_buffer[pos].src);
                        break;
                        
                    case RERR_TYPE:
                        /*****************************************************************/
                        /*                            RERR                               */
                        /*****************************************************************/	  
                        //IMPROVE - Right now we do not try local repair. This could be added.
                        ROUTING_DEBUG("\n<<<RERR from %u",aodv_recv_control_packet_buffer[pos].src);
                        aodv_rerr_recv_count++;
                        
                        //Set led if used
                        if(aodv_use_led)
                            btn_led_set(LED1);
                            
                            //If affected, deactivate routing table entries and forward RERR
                            
                            //Prepare RERR
                            aodv_control_rerr_packet_t rerr_in;
                        aodv_pkt_deserialize_rerr(&rerr_in, aodv_recv_control_packet_buffer[pos].pkt->data);
                        
                        aodv_control_rerr_packet_t rerr_out;
                        rerr_out.type = RERR_TYPE;
                        rerr_out.flags = 0;
                        rerr_out.destcount = 0;
                        
                        //Deactivate any entry in routing table that utilizes next hop entry
                        //Also register the destinations for the rerr
                        //IMPROVE - Dynamically allocating RERRs that allow to care more than only
                        //          two unreachable destination would result in less traffic. However,
                        //          the code gets more complicated and overhead for dynamic memory management
                        //          increases.
                        NutMutexLock(&mutex_routing_table);
                        for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
                            //If this route has the node with the route error next hop, and the destionation
                            //matches an entry in the route error, then deactivate.
                            if(routing_table[i].flags & FLAG_ACTIVE &&
                               routing_table[i].next_hop_address == aodv_recv_control_packet_buffer[pos].src &&
                               (routing_table[i].dst_address == rerr_in.unreachable_destination_address ||
                                (rerr_in.destcount == 2 && 
                                 routing_table[i].dst_address == rerr_in.additional_unreachable_destination_address))) {
#ifdef ROUTING_INFOMODE
                                ROUTING_DEBUG("\nI: (aodv_recv_thread) Deactivating routing table entry %u.\n",i);
#endif //ROUTING_INFOMODE
                                routing_table[i].flags &= (~FLAG_ACTIVE);
                                routing_table[i].lifetime = NutGetMillis() + DELETE_PERIOD;
                                if(rerr_out.destcount == 0) {
                                    //Register unreachable destination
                                    rerr_out.unreachable_destination_address = routing_table[i].dst_address;
                                    rerr_out.unreachable_dsn = routing_table[i].dsn;
                                    rerr_out.destcount++;
                                } else if(rerr_out.destcount == 1) {
                                    //Register addtional unreachable destination
                                    rerr_out.additional_unreachable_destination_address = routing_table[i].dst_address;
                                    rerr_out.additional_unreachable_dsn = routing_table[i].dsn;
                                    
                                    //Send RERR, since it is full
                                    aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr_out);
                                    
                                    //Reset RERR, so that a new RERR is sent if aditional destinations exist.  
                                    rerr_out.flags = 0;
                                    rerr_out.destcount = 0;
                                } 
                            }
                        }
                            NutMutexUnlock(&mutex_routing_table);  
                        
                        //If the count is one, the RERR_out must be sent
                        if(rerr_out.destcount == 1) {
                            aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr_out);
                        }
                            
                            //Clear led if used
                            if(aodv_use_led)
                                btn_led_clear(LED1);
                            
                            break;
                        
                    default:
                        ROUTING_DEBUG("\n<<< Invalid Packet, dropping!");
                        aodv_invalid_count++;
	}// end switch
                
                //Free this packet and mark slot empty
#ifdef ROUTING_DEBUGMODE
                if(aodv_recv_control_packet_buffer[pos].pkt == NULL)
                    ROUTING_DEBUG("\nA: (recv_control_thread) Going to free buffer position that is already NULL.\n");
#endif //ROUTING_DEBUGMODE
                free(aodv_recv_control_packet_buffer[pos].pkt);
                aodv_recv_control_packet_buffer[pos].pkt = NULL;
                aodv_recv_control_packet_buffer[pos].state = FREE;
                
                //Decrease counter
                aodv_recv_control_packet_buffer_counter--;
      } //end else
            
            //Increase pos to next position
            //pos = (pos + 1) % CONTROL_PACKET_QUEUE_SIZE;
            
    } //end else packets in queue waiting
        
        //IMPROVE - Probably one wants to yield here and give other threads a chance. It MUST be evaluated
        //          if this would increase or decrease performance.
  	//printf("THREAD recv_control_thread (counter=%u,packets=%u)\n", aodv_recv_control_packet_buffer_counter, packet_count);        
  }
}


/**
* Handler for incomming packets
 **/
static void control_handler(ccc_packet_t *pkt)
{
    //static u_short pos = 0;
    u_short i;
    
    // Check if there is enough space in the ringbuffer (queue)
    if(aodv_recv_control_packet_buffer_counter == CONTROL_PACKET_QUEUE_SIZE) {
#ifdef ROUTING_DEBUGMODE
        fflush(stdout); printf("\nE: (control_handler) Buffer full!\n"); fflush(stdout);
#endif //ROUTING_DEBUGMODE
        return;
    } else {
        //Search for free position
        for(i = 0; i < CONTROL_PACKET_QUEUE_SIZE; i++) {
            if(aodv_recv_control_packet_buffer[i].pkt == NULL && aodv_recv_control_packet_buffer[i].state == FREE)
                break;
#ifdef ROUTING_DEBUGMODE
            else if(aodv_recv_control_packet_buffer[i].state == FREE && aodv_recv_control_packet_buffer[i].pkt != NULL) {
                //Shouldn't happen, really!
                fflush(stdout); printf("\nA: (control_handler) Memory leak posibble...\n"); fflush(stdout);
            }
#endif //ROUTING_DEBUGMODE
        }
#ifdef ROUTING_DEBUGMODE
        if(i == CONTROL_PACKET_QUEUE_SIZE) {
            //Shouldn't happen, really!
            fflush(stdout); printf("\nA: (control_handler) Counter desync...\n"); fflush(stdout);
        }
#endif //ROUTING_DEBUG
    }
    
    // Reserve space for packet
    if( (aodv_recv_control_packet_buffer[i].pkt = malloc(sizeof(ccc_packet_t) + pkt->length)) != 0 ) {
        //Success
        memcpy(aodv_recv_control_packet_buffer[i].pkt,pkt,(sizeof(ccc_packet_t) + pkt->length));
        aodv_recv_control_packet_buffer[i].src = pkt->src;
        aodv_recv_control_packet_buffer[i].state = FULL;
        
        //Increase number of packets in queue
        aodv_recv_control_packet_buffer_counter++;
        
        //Signal that there is a new packet, just in case the receiver thread sleeps.
        //Do this async, since this routine runs in interupt context.
        NutEventPostAsync(&event_control_packet_recvd);
    } else { 
        // Space allocation failed, drop packet (==abort)
#ifdef ROUTING_DEBUGMODE
        fflush(stdout); printf("\nE: (control_handler) Not enough space for packet.\n"); fflush(stdout);
#endif //ROUTING_DEBUGMODE
    }
}


//------------------------------
// Route Maintance
//------------------------------
THREAD(aodv_hello_thread,arg)
{
    aodv_control_rrep_packet_t rrep;
    int i;
    
    //Define Hello Message
    rrep.type = RREP_TYPE;
    rrep.flags = 0;                //Do not ACK Hello messages
    rrep.hopcount = 0;
    rrep.destination_address = aodv_get_address();
    rrep.dsn = aodv_sequence_no;
    rrep.originator_address = BROADCAST_ADDR;
    rrep.lifetime = ALLOWED_HELLO_LOSS * HELLO_INTERVAL;
    rrep.ttl = 1;
    
    for(;;) {
        
        //printf("THREAD aodv_hello_thread\n");
        for(i = 0; i < ROUTING_TABLE_SIZE; i++)
            if(rt_check_entry_activity(i))
                break;
        
        if(i != ROUTING_TABLE_SIZE &&                           //There is an active route
           NutGetMillis() - last_broadcast > HELLO_INTERVAL) {
            //No broadcast in last intervall - Send hello Message
            
            //Increase sequence no
            //IMPROVE - Discuss: Not specified in the RFC, however, since this is a rrep
            //and according to the RFC any RREP increases the sequence number, let's do
            //this (doesn't make much sens otherwise)
            aodv_sequence_no++;
            rrep.dsn = aodv_sequence_no;
            
            aodv_send_control_rrep_packet(BROADCAST_ADDR, &rrep);
        }
        NutSleep(HELLO_INTERVAL);
    }
}


//------------------------------
// Control Packets Management
//------------------------------

/**
* Creates a new AODV control packet
 *
 * @param type [in] packet type
 * @return pointer to packet
 */
inline static void *new_aodv_control_packet(int type)
{
    switch(type) {
        case RREQ_TYPE:
            return malloc(sizeof(aodv_control_rreq_packet_t));
            break;
            
        case RREP_TYPE:
            return malloc(sizeof(aodv_control_rrep_packet_t));
            break;
            
        case RREP_ACK_TYPE:
            return malloc(sizeof(aodv_control_rrep_ack_packet_t));
            break;
            
        case RERR_TYPE:
            return malloc(sizeof(aodv_control_rerr_packet_t));
            break;
            
        default:
            return NULL;
    }
}

/**
* Serialize AODV control packet to char buffer 
 *
 * Used for sending the packet as payload in a ccc_packet
 */  
static inline void aodv_pkt_serialize_rreq(u_char *to, aodv_control_rreq_packet_t *from)
{
    to[0] = from->type;                                                //char
    to[1] = from->flags;                                               //char
    to[2] = from->hopcount;                                            //char
    to[3] = from->reserved;                                            //char
    *((u_short*) &to[4]) = htons((u_short) from->id);                  //short
    *((u_short*) &to[6]) = htons((u_short) from->destination_address); //short
    *((u_short*) &to[8]) = htons((u_short) from->dsn);                 //short
    *((u_short*) &to[10]) = htons((u_short) from->originator_address); //short
    *((u_short*) &to[12]) = htons((u_short) from->osn);                //short
    *((u_short*) &to[14]) = htons((u_short) from->ttl);                //short
}

static inline void aodv_pkt_deserialize_rreq(aodv_control_rreq_packet_t *to, u_char *from)
{
    to->type = from[0];                                     //char
    to->flags = from[1];                                    //char
    to->hopcount = from[2];                                 //char
    to->reserved = from[3];                                 //char
    to->id = ntohs(*((u_short*)&from[4]));                  //short
    to->destination_address = ntohs(*((u_short*)&from[6])); //short
    to->dsn = ntohs(*((u_short*)&from[8]));                 //short
    to->originator_address = ntohs(*((u_short*)&from[10])); //short
    to->osn = ntohs(*((u_short*)&from[12]));                //short
    to->ttl = ntohs(*((u_short*)&from[14]));                //short
}

static inline void aodv_pkt_serialize_rrep(u_char *to, aodv_control_rrep_packet_t *from)
{
    to[0] = from->type;                                                //char
    to[1] = from->flags;                                               //char
    to[2] = from->prefix_sz;                                           //char
    to[3] = from->hopcount;                                            //char
    *((u_short*) &to[4]) = htons((u_short) from->destination_address); //short
    *((u_short*) &to[6]) = htons((u_short) from->dsn);                 //short
    *((u_short*) &to[8]) = htons((u_short) from->originator_address);  //short
    *((u_long*)  &to[10]) = htonl((u_long) from->lifetime);            //long
    *((u_short*) &to[14]) = htons((u_short) from->ttl);                //short
}

static inline void aodv_pkt_deserialize_rrep(aodv_control_rrep_packet_t *to, u_char *from)
{
    to->type = from[0];                                     //char
    to->flags = from[1];                                    //char
    to->prefix_sz = from[2];                                //char
    to->hopcount = from[3];                                 //char
    to->destination_address = ntohs(*((u_short*)&from[4])); //short
    to->dsn = ntohs(*((u_short*)&from[6]));                 //short
    to->originator_address = ntohs(*((u_short*)&from[8]));  //short
    to->lifetime = ntohl(*((u_long*)&from[10]));            //long
    to->ttl = ntohs(*((u_short*)&from[14]));                //short
}

static inline void aodv_pkt_serialize_rrep_ack(u_char *to, aodv_control_rrep_ack_packet_t *from)
{
    to[0] = from->type;                                      //char
    to[1] = from->reserved;                                  //char
}

static inline void aodv_pkt_deserialize_rrep_ack(aodv_control_rrep_ack_packet_t *to, u_char *from)
{
    to->type     = from[0];                                  //char
    to->reserved = from[1];                                  //char
}

static inline void aodv_pkt_serialize_rerr(u_char *to, aodv_control_rerr_packet_t *from)
{
    to[0] = from->type;                                                                       //char
    *((u_short*) &to[1]) = htons((u_short) from->flags);                                      //short
    to[3] = from->destcount;                                                                  //char
    *((u_short*) &to[4]) = htons((u_short) from->unreachable_destination_address);            //short
    *((u_short*) &to[6]) = htons((u_short) from->unreachable_dsn);                            //short
    *((u_short*) &to[8]) = htons((u_short) from->additional_unreachable_destination_address); //short
    *((u_short*) &to[10]) = htons((u_short) from->additional_unreachable_dsn);                //short
}

static inline void aodv_pkt_deserialize_rerr(aodv_control_rerr_packet_t *to, u_char *from)
{
    to->type     = from[0];                                                        //char
    to->flags    = ntohs(*((u_short*)&from[1]));                                   //short
    to->destcount = from[3];                                                       //char
    to->unreachable_destination_address = ntohs(*((u_short*)&from[4]));            //short
    to->unreachable_dsn = ntohs(*((u_short*)&from[6]));                //short
    to->additional_unreachable_destination_address = ntohs(*((u_short*)&from[8])); //short
    to->additional_unreachable_dsn = ntohs(*((u_short*)&from[10]));    //short
}


/////////////////////////////////
// Buffer Management
/////////////////////////////////
static char rreq_buffer_add(u_short id, u_short originator_address)
{
    int i;
    
    //Aquire mutex on shared ressource
    NutMutexLock(&mutex_rreq_buffer);
    
    //Search open position
    for(i=0; i < RREQ_BUFFER_SIZE; i++) {
        if((long)rreq_buffer[i].timestamp < (long)(NutGetMillis() - PATH_DISCOVERY_TIME)) { //IMPROVE - change overflow behav.
            rreq_buffer[i].id = id;
            rreq_buffer[i].originator_address = originator_address;
            rreq_buffer[i].timestamp = NutGetMillis();
            break;
        }
    }
    //Release mutex
    NutMutexUnlock(&mutex_rreq_buffer);
    
    if(i == RREQ_BUFFER_SIZE) {
        //No more free space, return error
        ROUTING_DEBUG("Cannot add RREQ to buffer: not enough space\n");
        return -1;
    } else {
        //Success
        return 0;
    }
}

static char rreq_buffer_lookup(u_short id, u_short originator_address)
{
    int i;
    
    //Search entry
    NutMutexLock(&mutex_rreq_buffer); //Aquire mutex on shared ressource
    for(i=0; i < RREQ_BUFFER_SIZE; i++) {
        if((rreq_buffer[i].id == id) &&
           (rreq_buffer[i].originator_address == originator_address) &&
           (rreq_buffer[i].timestamp >= NutGetMillis() - PATH_DISCOVERY_TIME) ) {
            break;
        }
    }
    NutMutexUnlock(&mutex_rreq_buffer); //Release mutex
    
    if(i == RREQ_BUFFER_SIZE) {
        //No entry found
        return -1;
    } else {
        //Entry found
        return 0;
    }
}

static char rrep_buffer_add(aodv_control_rrep_packet_t *rrep, u_short dst)
{
    int i;
    
    //Aquire mutex on shared ressource
    NutMutexLock(&mutex_rrep_buffer);
    
    //Search open position
    for(i=0; i < RREP_BUFFER_SIZE; i++) {
        if(aodv_rrep_buffer[i].rrep == NULL) { 
            if( (aodv_rrep_buffer[i].rrep = new_aodv_control_packet(RREP_TYPE)) == NULL) {
                ROUTING_DEBUG("\nE: (rrep_buffer_add) Space allocation for rreq failed.\n");
                NutMutexUnlock(&mutex_rrep_buffer);
                return -1;
            } else {
                memcpy(aodv_rrep_buffer[i].rrep, rrep, sizeof(aodv_control_rrep_packet_t));
                aodv_rrep_buffer[i].dst = dst;
                aodv_rrep_buffer[i].retry = 0;
                aodv_rrep_buffer[i].nextshd = 0;
                break;
            }
        }
    }
    //Release mutex
    NutMutexUnlock(&mutex_rrep_buffer);
    
    if(i == RREP_BUFFER_SIZE) {
        //No more free space, return error
        ROUTING_DEBUG("\nE: (rrep_buffer_add) Not enough space in buffer!\n");
        return -1;
    } else {
        //Success
        NutMutexLock(&mutex_rrep_buffer_count);
        aodv_rrep_buffer_count++;
        NutMutexUnlock(&mutex_rrep_buffer_count);
        return 0;
    }
}

/**
* Removes ANY rrep pending for dst
 **/
static char rrep_buffer_del(u_short dst)
{
    int  i;
    char count = 0;
    
    //Aquire mutex on shared ressource
    NutMutexLock(&mutex_rrep_buffer);
    
    //Search entry
    for(i=0; i < RREP_BUFFER_SIZE; i++) {
        if(aodv_rrep_buffer[i].rrep != NULL && aodv_rrep_buffer[i].dst == dst) { 
            //Entry for destination found, remove
            free(aodv_rrep_buffer[i].rrep);
            aodv_rrep_buffer[i].rrep = NULL;
            aodv_rrep_buffer[i].dst = 0;
            aodv_rrep_buffer[i].retry = 0;
            aodv_rrep_buffer[i].nextshd = 0;
            NutMutexLock(&mutex_rrep_buffer_count);
            aodv_rrep_buffer_count--;
            NutMutexUnlock(&mutex_rrep_buffer_count);
            count++;
        }
    }
    //Release mutex
    NutMutexUnlock(&mutex_rrep_buffer);
    
    //Return number of deleted packets
    return count;
}


/**
* Adds a new rreq to the discovery buffer.
 *
 */
static inline char rreq_discovery_buffer_add(aodv_control_rreq_packet_t *rreq)
{
    int j;
    
    NutMutexLock(&mutex_discovery_buffer);
    
    for(j = 0; j < DISCOVERY_BUFFER_SIZE; j++) {
        if(aodv_discovery_buffer[j].rreq == NULL)
            break;
    }
    
    if(j == DISCOVERY_BUFFER_SIZE) {
        //No more free space in discovery buffer
        ROUTING_DEBUG("\nE: (rreq_discovery_buffer_add) No more space in discovery buffer.\n");
        NutMutexUnlock(&mutex_discovery_buffer);
        return -1;
    }
    
    aodv_discovery_buffer[j].retry = 0;
    aodv_discovery_buffer[j].nextshd = 0;
    aodv_discovery_buffer[j].rreq = rreq;
    
    NutMutexLock(&mutex_discovery_buffer_count);
    aodv_discovery_buffer_count++;
    NutMutexUnlock(&mutex_discovery_buffer_count);
    
    NutMutexUnlock(&mutex_discovery_buffer);
    
    return j;
}

/**
* Adds a new entry to the blacklist.
 *
 * @return 0 on success, -1 otherwise.
 */
static inline char aodv_blacklist_add(u_short address)
{
    int i;
    for(i = 0; i < BLACKLIST_SIZE; i++) {
        if(NutGetMillis() > blacklist[i].lifetime) {
            //This is a free slot
            break;
        }
    }
    
    if(i < BLACKLIST_SIZE) {
        blacklist[i].address = address;
        blacklist[i].lifetime = NutGetMillis() + (long)BLACKLIST_TIMEOUT;
        return 0;
    } else {
        //No more space in blacklist, ignore
        ROUTING_DEBUG("\nE: (aodv_blacklist_add) No more space in blacklist!\n");
        return -1;
    } 
}

/**
* Checks wheter a node is blacklisted or not
 *
 * @return 1 if node is blacklisted, 0 otherwise.
 */
static inline u_short aodv_blacklist_lookup(u_short address)
{
    int i;
    for(i = 0; i < BLACKLIST_SIZE; i++) {
        if(blacklist[i].address == address && blacklist[i].lifetime > NutGetMillis()) 
            return 1;
    }
    return 0;
}


/////////////////////////////////
// Routing Table Management 
////////////////////////////////

//IMPROVE - Check wheter supplied index is inside boundaries 
//          to be on the save side. I decided to safe this check (performance++).

/**
* Checks if a routing table entry is valid or not
 *
 * @param [in} entry number
* @return 1 if valid, 0 otherwise
*/
inline static u_char rt_check_entry_validity(u_short i)
{
    char result;
    
    NutMutexLock(&mutex_routing_table);
    result = routing_table[i].flags & FLAG_VALID;
    NutMutexUnlock(&mutex_routing_table);
    
    return result;
}

/**
* Checks if a routing table entry is active or not
 *
 * @param [in} entry number
* @return 1 if active, 0 otherwise
*/
inline static u_char rt_check_entry_activity(u_short i)
{
    char result;
    
    NutMutexLock(&mutex_routing_table);
    result = routing_table[i].flags & FLAG_ACTIVE;
    NutMutexUnlock(&mutex_routing_table);
    
    //Return the flag
    return result;
}


/**
* Returns the entry that contains the information for rouing
 * towards a destination.
 * 
 * @param [in] destination address
 * @return entry position, -1 on error
 */
inline static short rt_create_entry(u_short dst, u_short dsn, u_short flags, 
                                    u_short metric, u_short next_hop, int lifetime)
{
    u_short i;
    
    // Search for free slot
    NutMutexLock(&mutex_routing_table);   //Get Mutex since we do not want the table changing during search
    for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
        if(routing_table[i].lifetime + DELETE_PERIOD < NutGetMillis())
            break;
    }
    NutMutexUnlock(&mutex_routing_table);
    
    //If an empty slot is found make the entry, otherwise return -1
    if(i < ROUTING_TABLE_SIZE) {
        NutMutexLock(&mutex_routing_table);
        routing_table[i].dst_address = dst;
        routing_table[i].dsn = dsn;
        routing_table[i].flags = flags;
        routing_table[i].metric = metric;
        routing_table[i].next_hop_address = next_hop;
        routing_table[i].lifetime = NutGetMillis() + lifetime;
        NutMutexUnlock(&mutex_routing_table);
        return i;
    } else {
        ROUTING_DEBUG("\nE: (rt_create_entry) Failed adding route.\n");
        return -1;
    }
}

inline static char rt_update_entry(short handle)
{
    //If i is in range make update, otherwise return -1
    if(handle >= 0 && handle < ROUTING_TABLE_SIZE) {
        NutMutexLock(&mutex_routing_table);
        routing_table[handle].lifetime = UMAX(routing_table[handle].lifetime, (NutGetMillis() + ACTIVE_ROUTE_TIMEOUT));
        NutMutexUnlock(&mutex_routing_table);
        return 0;
    } else {
        ROUTING_DEBUG("\nE: (rt_update_entry) Failed updating route.\n");
        return -1;
    }
}

inline static void rt_deactivate_next_hop(u_short next_hop)
{
    short i;
    
    // Search for matches
    NutMutexLock(&mutex_routing_table);
    for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
        //If this route has the node with the link break as next hop, deactivate.
        if(routing_table[i].next_hop_address == next_hop) {
#ifdef ROUTING_INFOMODE
            ROUTING_DEBUG("\nI: (rt_get_entry) Deactivating routing table entry %u.\n",i);
#endif //ROUTING_INFOMODE
            routing_table[i].flags &= (~FLAG_ACTIVE);
            routing_table[i].lifetime = (u_long)DELETE_PERIOD;
        }
    }
    NutMutexUnlock(&mutex_routing_table);  
}


/**
* Returns the entry that contains the information for rouing
 * towards a destination.
 * 
 * @param [in] destination address
 * @return entry position, -1 on error
 */
inline static short rt_get_entry(u_short dst) 
{  
    short i;
    
    // Look up destination in Table
    NutMutexLock(&mutex_routing_table);   //Get Mutex since we do not want the table changing during search
    for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
        //Check if route entry timed out (Lazy)
        if(routing_table[i].flags & FLAG_ACTIVE) { 
            if(routing_table[i].lifetime < NutGetMillis()) {//Note: Route may be inactive even though not timed out. 
                                                            //Route timed out, deacivate
#ifdef ROUTING_INFOMODE
                ROUTING_DEBUG("\nI: (rt_get_entry) Deactivating routing table entry %u.\n",i);
#endif //ROUTING_INFOMODE
                routing_table[i].flags &= (~FLAG_ACTIVE);
                routing_table[i].lifetime += DELETE_PERIOD;
            }
        }
        
        //If entry found, break
        if(routing_table[i].dst_address == dst)
            break;
    }
    NutMutexUnlock(&mutex_routing_table);
    
    //If an entry is found return its position, otherwise return -1
    if(i < ROUTING_TABLE_SIZE) {
        return i;
    } else {
        return -1;
    }
}


/**
* Returns the selected Flags
 **/
inline static u_short rt_get_flags(u_short entry, u_short flags)
{
    u_short result;
    
    NutMutexLock(&mutex_routing_table);
    result = routing_table[entry].flags & flags;
    NutMutexUnlock(&mutex_routing_table);
    
    return result;
}

/**
* Returns the sequence number of the selected entry
 **/
inline static u_short rt_get_sequence_no(u_short entry)
{
    u_short result;
    
    NutMutexLock(&mutex_routing_table);
    result = routing_table[entry].dsn;
    NutMutexUnlock(&mutex_routing_table);
    
    return result;
}


////////////////////////////////////////////////////
// Debug Stuff
////////////////////////////////////////////////////
//
// Only allowed when debuging, since these functions
// allow to mess up pretty everything
//
#ifdef ROUTING_DEBUGMODE

int aodv_debug_send_rreq_packet(u_char flags, u_char hopcount,
                                u_char reserved, u_short id, u_short destination_address,
                                u_short dsn, u_short originator_address, u_short osn,
                                u_short ttl) 
{
    int res;
    aodv_control_rreq_packet_t *rreq;
    
    //Allocate Packet
    if((rreq = new_aodv_control_packet(RREQ_TYPE)) == NULL) {
        //Packet allocation failed, return with error
        ROUTING_DEBUG("\nE: (aodv_debug_send_rreq_packet) Space allocation for RREQ failed.\n");
        return -1;
    }  
    
    //Fill in Packet Data
    rreq->type = RREQ_TYPE;
    rreq->flags = flags;
    rreq->hopcount = hopcount;
    rreq->reserved = reserved; 
    rreq->id = id;
    rreq->destination_address = destination_address;
    rreq->dsn = dsn;
    rreq->originator_address = originator_address;
    rreq->osn = osn;
    rreq->ttl = ttl;
    
    
    //Send the packet
    res = aodv_send_control_rreq_packet(BROADCAST_ADDR, rreq);
    
    //Free memory
    free(rreq);
    
    return res;
}

int aodv_debug_send_rrep_packet(u_short addr) 
{
    aodv_control_rrep_packet_t *rrep;
    
    //Allocate Packet
    if((rrep = new_aodv_control_packet(RREP_TYPE)) == NULL) {
        //Packet allocation failed, return with error
        ROUTING_DEBUG("\nE: (aodv_debug_send_rrep_packet) Space allocation for RREP failed.\n");
        return -1;
    } 
    
    //Fill the Packet
    return aodv_send_control_rrep_packet(addr, rrep);
}

int aodv_debug_send_rrep_ack_packet(u_short addr) 
{
    aodv_control_rrep_ack_packet_t *rrep_ack;
    
    //Allocate the Packet
    if((rrep_ack = new_aodv_control_packet(RREP_ACK_TYPE)) == NULL) {
        //Packet allocation failed, return with error
        ROUTING_DEBUG("\nE: (aodv_debug_send_rrep_ack_packet) Space allocation for RREP_ACK failed.\n");
        return -1;
    } 
    
    //Fill the packet
    
    return aodv_send_control_rrep_ack_packet(addr, rrep_ack);
}


void aodv_set_sequence_no(int seqno) 
{
    aodv_sequence_no = seqno;
}


void aodv_set_rreq_id(int rreq_id)
{
    aodv_rreq_id = rreq_id;
}


#endif //ROUTING_DEBUGMODE


////////////////////////////////
// Setters & Getters
///////////////////////////////

u_short aodv_get_address(void)
{
    return aodv_address;
}

u_short aodv_set_address(u_short addr)
{
    aodv_address = addr;
    return 0;
}


/////////////////////////////////////////////////////////////////////////////////
// Testing
/////////////////////////////////////////////////////////////////////////////////
// Functions that allow testing, but do not add any functionality to the protocol
// IMPROVE - Add switch, such as ROUTING_DEBUGMODE to deactivate testing commands

int aodv_route_discover(u_short dst) {
    
    int i,j;
    aodv_control_rreq_packet_t *rreq = NULL;
    
    //Assemble rreq
    if((rreq = new_aodv_control_packet(RREQ_TYPE)) == NULL) {
        //Packet allocation failed, return with error
        ROUTING_DEBUG("\nE: (aodv_route_discover) Space allocation for RREQ failed.\n");
        return -1;
    } 
    rreq->flags &= 0; 
    rreq->id = ++aodv_rreq_id; //Every time a RREQ is emmited (also resend) id is incremented (Page 15)
    rreq->hopcount = 0; 
    rreq->type = RREQ_TYPE; 
    
    //Register adresses 
    rreq->destination_address = dst; 
    rreq->originator_address = aodv_get_address(); 
    
    //Increment sequence number 
    /* This prevents conflicts with previously established 
        reverse routes.*/
    aodv_sequence_no++; 
    
    //Set originator (my own) sequence number 
    /* This must be done after the sequence number is increased */
    rreq->osn = aodv_sequence_no; 
    
    //Set the destination sequence number field in the rreq package. 
    /* The destination sequence number field in the RREQ  
        message is the last known destination sequence number 
        for this destination and is copied from the 
        destination sequence number field in the routing  
        table. If no sequence number is nown, the unknown 
        sequence number flag MUST be set. */
    
    if((i = rt_get_entry(dst)) != -1) {
        //Invalid entry exists - clear flag and store sequence no
        rreq->flags &= (~FLAG_UNKNOWN_SEQUENCE_NUMBER);
        rreq->dsn = rt_get_sequence_no(i); //
                                           //IMPROVE - Set ttl for expanding ring search
                                           //rreq->ttl = UMAX(rt_get_hopcount(i) + TTL_INCREMENT, NET_DIAMETER);
        rreq->ttl = 0xFFFF;
    } else {
        //Entry does not exist, set unknown sequence no flag
        rreq->flags |= FLAG_UNKNOWN_SEQUENCE_NUMBER;
        //Set ttl for expanding ring search
        rreq->ttl = 0xFFFF; 
        //IMPROVE - Code for expanding ring search
    }
    
    //Add the route request to the discovery_buffer
    NutMutexLock(&mutex_discovery_buffer);
    NutMutexLock(&mutex_discovery_buffer_count);
    for(j=0; j<DISCOVERY_BUFFER_SIZE; j++) {
        if(aodv_discovery_buffer[j].rreq == NULL)
            break;
    }
    if(j == DISCOVERY_BUFFER_SIZE) {
        //No more free space in discovery buffer
        ROUTING_DEBUG("\nE: (aodv_route_discover) No more space in discovery buffer.\n");
        NutMutexUnlock(&mutex_discovery_buffer_count);
        NutMutexUnlock(&mutex_discovery_buffer);
        return 0;
    }
    
    aodv_discovery_buffer[j].retry = 0;
    aodv_discovery_buffer[j].nextshd = 0;
    aodv_discovery_buffer[j].rreq = rreq;
    
    aodv_discovery_buffer_count++;
    
    NutMutexUnlock(&mutex_discovery_buffer_count);
    NutMutexUnlock(&mutex_discovery_buffer);
    
    //Message upper layer that the route discovery will be started
    //(if a message handler was installed)
    if(aodv_message_handler != NULL)
        aodv_message_handler(dst, DISCOVERY_STARTED);
    
    //Start the discovery
    NutEventPost(&event_rreq_pending);
    
    return 0;
}

int aodv_get_sequence_no(void) {
    return aodv_sequence_no;
}

int aodv_get_rreq_id(void) {
    return aodv_rreq_id;
}

void aodv_print_rt(void) {
    int i; u_long curtime;
    u_long th,tm,ts,tms;
    u_short h,m,s,ms;
    
    printf("No. DST/  DSN V A Metric NextHop Lifetime\n");
    
    for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
    	
    	// only display entries that contain a route
    	if (routing_table[i].dst_address == 0) continue;
    	
        //Calculate lifetime
        curtime = NutGetMillis();
        if(routing_table[i].lifetime != 0 && routing_table[i].lifetime > curtime) {
            
            th = (routing_table[i].lifetime - curtime) / 1000;
            th /= 60; th/= 60;
            h = (u_short)th;
            
            tm = (routing_table[i].lifetime - curtime);
            th *= 1000; th *= 60; th *= 60;
            tm -= th; 
            tm /= 1000; tm /= 60;
            m = (u_short)tm;
            
            ts = (routing_table[i].lifetime - curtime);
            tm *= 1000; tm *= 60;
            ts -= th; ts -= tm;
            ts /= 1000;
            s = (u_short)ts;
            
            tms = (routing_table[i].lifetime - curtime);
            ts *= 1000;
            tms -= th; tms -= tm; tms -= ts;
            ms = (u_short)tms;
        } else {
            h = 0;
            m = 0;
            s = 0;
            ms = 0;
        }
        printf("%2u  %3u/%5u %c %c %6u %7u %3u:%2u:%2u,%3u\n", i, routing_table[i].dst_address, routing_table[i].dsn,
               routing_table[i].flags & FLAG_VALID ? 'V' : '-' , 
               ((routing_table[i].flags & FLAG_ACTIVE) &&  (routing_table[i].lifetime > NutGetMillis())) ? 'A' : '-',
               routing_table[i].metric, routing_table[i].next_hop_address, h, m, s, ms
               );
    }
} 

// print the routing table entry for destination dest
void aodv_print_rt_entry(int dest) {
	
	int i = rt_get_entry(dest);
	if ( i < 0 )
	{
		printf( "No route to 0x%04X.\n", dest );
		return;
	}
	
    if(i >= 0 && i < ROUTING_TABLE_SIZE) {
        printf("Entry %u at %ldms:\n", i, NutGetMillis());
        printf("- Destination Address:      %u\n", routing_table[i].dst_address);
        printf("- Destination Sequence No.: %u\n", routing_table[i].dsn);
        printf("- Flags: Valid DSN:         %u\n", routing_table[i].flags & FLAG_VALID);
        printf("-        Active:            %u\n", routing_table[i].flags & FLAG_ACTIVE);
        printf("- Metric:                   %u\n", routing_table[i].metric);
        printf("- Next Hop Address:         %u\n", routing_table[i].next_hop_address);
        printf("- Lifetime:                 %ld\n\n", routing_table[i].lifetime);
    }
} 

////////////////////////////////
// Messaging
///////////////////////////////
int aodv_reg_msg_handler(routing_message_handler_t message_handler) 
{
    if( aodv_message_handler != NULL )
        return -1;
    
    aodv_message_handler = message_handler;
    return 0;
}

void aodv_recv_msg(u_short msg, u_short param) 
{
    aodv_control_rerr_packet_t rerr;
    int i;
    
    switch(msg) {
        case BROKEN_LINK:
            //IMPROVE - Create RERR dinamically to allow more than just two entries. This
            //          would save network utilisation. Also improve the local repair by
            //          doing it here as specified in Request for Comments 3561! Currently
            //          this is done a little bit lazy by accepting that the RFC layer
            //          tries to repair routes.
            
            //Prepare RERR
            rerr.type = RERR_TYPE;
            rerr.flags = 0;
            rerr.destcount = 0;
            
            //Deactivate any entry in routing table that utilizes next hop entry
            //Also register the destinations for the rerr.
            NutMutexLock(&mutex_routing_table);
            for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
                //If this route has the node with the link break as next hop, deactivate.
                if(routing_table[i].flags & FLAG_ACTIVE &&
                   routing_table[i].next_hop_address == param) {
#ifdef ROUTING_INFOMODE
                    ROUTING_DEBUG("\nI: (aodv_recv_msg) Deactivating routing table entry %u.\n",i);
#endif //ROUTING_INFOMODE
                    routing_table[i].flags &= (~FLAG_ACTIVE);
                    routing_table[i].lifetime = NutGetMillis() + DELETE_PERIOD;
                    if(rerr.destcount == 0) {
                        //Register unreachable destination
                        rerr.unreachable_destination_address = routing_table[i].dst_address;
                        rerr.unreachable_dsn = routing_table[i].dsn;
                        rerr.destcount++;
                    } else if(rerr.destcount == 1) {
                        //Register addtional unreachable destination
                        rerr.additional_unreachable_destination_address = routing_table[i].dst_address;
                        rerr.additional_unreachable_dsn = routing_table[i].dsn;
                        
                        //Send RERR, since it is full
                        aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr);
                        
                        //Reset RERR, so that a new RERR is sent if aditional destinations exist.  
                        rerr.type = RERR_TYPE;
                        rerr.flags = 0;
                        rerr.destcount = 0;
                    } 
                }
            }
                NutMutexUnlock(&mutex_routing_table);  
            
            //If the count is one, this RERR must be sent
            if(rerr.destcount == 1) {
                aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr);
            }
                break;
            
            
        case BROKEN_ROUTE:
            //IMPROVE - Start local repair and message ROUTE_AVAILBLE if successfull. Local repair
            //          must be done as specified in Request for Comments 3561! Currently
            //          this only generates a RERR and message that repair failed.
            
            //Prepare RERR
            rerr.type = RERR_TYPE;
            rerr.flags = 0;
            rerr.destcount = 0;
            
            //Deactivate any entry in routing table that utilizes next hop entry
            //Also register the destinations for the rerr.
            NutMutexLock(&mutex_routing_table);
            for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
                //If this route has the node with the link break as next hop, deactivate.
                if(routing_table[i].next_hop_address == param) {
                    //Deactivate entry if it is not yet deactivated.
                    if(routing_table[i].flags & FLAG_ACTIVE) {
                        routing_table[i].flags &= (~FLAG_ACTIVE); 
                        routing_table[i].lifetime = NutGetMillis() + DELETE_PERIOD;
                    }
                    if(rerr.destcount == 0) {
                        //Register unreachable destination
                        rerr.unreachable_destination_address = param;
                        rerr.unreachable_dsn = routing_table[i].dsn;
                        rerr.destcount++;
                    } else if(rerr.destcount == 1) {
                        //Register addtional unreachable destination
                        rerr.additional_unreachable_destination_address = param;
                        rerr.additional_unreachable_dsn = routing_table[i].dsn;
                        
                        //Send RERR, since it is full
                        aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr);
                        
                        //Reset RERR, so that a new RERR is sent if aditional destinations exist.  
                        rerr.type = RERR_TYPE;
                        rerr.flags = 0;
                        rerr.destcount = 0;
                    } 
                }
            }
                NutMutexUnlock(&mutex_routing_table);  
            
            //If the count is one, this RERR must be sent
            if(rerr.destcount == 1) {
                aodv_send_control_rerr_packet(BROADCAST_ADDR, &rerr);
            }
                
                //Send message to RFC
                if(aodv_message_handler != NULL)
                    aodv_message_handler(param, DESTINATION_UNREACHABLE);
                
                break;
            
            
        default:
            ROUTING_DEBUG("\nE: (aodv_recv_msg) Unknown message.\n");
    }
}

////////////////////////////////
// Misc
////////////////////////////////
void aodv_enable_led(int enable_led)
{
    aodv_use_led = (enable_led) ? 1 : 0;
}


////////////////////////////////
// Initialisation 
////////////////////////////////

/**
* Initializes all structures and Variables and also creates queue control
 * threads.
 */ 
int aodv_init(u_short address) 
{
    u_short i;
    long initial_wait;
    
    ROUTING_DEBUG("\n\n<-------- Initializing AODV... -------->\n\n ");
    
    //Initialize Timer
    NutTimerInit();
    
    //Block for ACTIVE_ROUTE_TIMEOUT
    btn_led_set(LED1);
    initial_wait = NutGetMillis() + ACTIVE_ROUTE_TIMEOUT;
    while(NutGetMillis() < initial_wait);
    btn_led_clear(LED1);
    
    //Initialize variables
    aodv_address = address;
    aodv_sequence_no = 0;
    aodv_rreq_id = 0;
    use_ack = 1; //IMPROVE - Dynamic sensing if acks are necessary would be nice.
    aodv_message_handler = NULL;
    aodv_rreq_recv_count = 0;              
    aodv_rreq_send_count = 0;              
    aodv_rreq_recv_fail_count = 0;         
    aodv_rreq_send_fail_count = 0;         
    aodv_rreq_ignore_count = 0;            
    aodv_rrep_recv_count = 0;              
    aodv_rrep_send_count = 0;              
    aodv_rrep_recv_fail_count = 0;         
    aodv_rrep_send_fail_count = 0;         
    aodv_rrep_ack_recv_count = 0;          
    aodv_rrep_ack_send_count = 0;          
    aodv_rrep_ack_send_fail_count = 0;     
    aodv_rerr_recv_count = 0;               
    aodv_rerr_send_count = 0;              
    aodv_rerr_send_fail_count = 0;         
    aodv_invalid_count = 0;                
    
    //Initialize Mutexes
    ROUTING_DEBUG_INIT;
    NutMutexInit(&mutex_routing_table);
    NutMutexInit(&mutex_rreq_buffer);
    NutMutexInit(&mutex_discovery_buffer);
    NutMutexInit(&mutex_discovery_buffer_count);
    NutMutexInit(&mutex_rrep_buffer);
    NutMutexInit(&mutex_rrep_buffer_count);
    
    aodv_recv_control_packet_buffer_counter = 0;
    aodv_discovery_buffer_count = 0;
    aodv_rrep_buffer_count = 0;
    
    //Initialize routing table
    NutMutexLock(&mutex_routing_table);
    for(i = 0; i < ROUTING_TABLE_SIZE; i++) {
        routing_table[i].dst_address = 0;
        routing_table[i].dsn = 0;
        routing_table[i].flags = 0;
        routing_table[i].metric = 0xFFFF;
        routing_table[i].next_hop_address = 0;
        routing_table[i].lifetime = 0;
    }
    NutMutexUnlock(&mutex_routing_table);
    
    
    //Initialize queues
    for(i = 0; i < CONTROL_PACKET_QUEUE_SIZE; i++) {
        aodv_recv_control_packet_buffer[i].pkt = NULL;
        aodv_recv_control_packet_buffer[i].state = FREE;
    }
    
    //Initialize Buffers
    for(i = 0; i < RREQ_BUFFER_SIZE; i++) {
        //Mark all entries invalid
        rreq_buffer[i].timestamp = -PATH_DISCOVERY_TIME; 
    }
    
    for(i = 0; i < DISCOVERY_BUFFER_SIZE; i++) {
        aodv_discovery_buffer[i].rreq = NULL;
        aodv_discovery_buffer[i].nextshd = 0;
        aodv_discovery_buffer[i].retry = 0;
    }
    
    for(i = 0; i < RREP_BUFFER_SIZE; i++) {
        aodv_rrep_buffer[i].rrep = NULL;
        aodv_rrep_buffer[i].retry = RREP_RETRY;
        aodv_rrep_buffer[i].nextshd = 0;
    }
    
    for(i = 0; i < BLACKLIST_SIZE; i++) {
        blacklist[i].address = 0;
        blacklist[i].lifetime = 0;
    }
    
    /* register ccc handlers */
    // Register Handler for control packets
    ccc_register_packet_handler(CONTROL_TYPE, control_handler);
    
    //Start threads
    if( 0 == NutThreadCreate("aodv_crec", recv_control_thread, 0 , 1024)) { //IMPROVE - Decide minimal stack size
                                                                            //Creating Thread failed, AODV will not operate
        ROUTING_DEBUG("\nE: (aodv_init) Create receive control thread failed!\n");
        return -1;
    }
    if( 0 == NutThreadCreate("aodv_rdsc", aodv_discover_thread, 0 , 1024)) { //IMPROVE - Decide minimal stack size
                                                                             //Creating Thread failed, AODV will not operate
        ROUTING_DEBUG("\nE: (aodv_init) Create route discover thread failed!\n");
        return -1;
    }
    if( 0 == NutThreadCreate("aodv_rrep", aodv_rrep_thread, 0 , 1024)) { //IMPROVE - Decide minimal stack size
                                                                         //Creating Thread failed, AODV will not operate
        ROUTING_DEBUG("\nE: (aodv_init) Create rrep thread!\n");
        return -1;
    }
    //If HELLO_INTERVAL is set, create the thread, otherwise skip it.
    //IMPROVE - Being able to dynamically adjust the hello interval would be nice.
    if( HELLO_INTERVAL != 0 && 0 == NutThreadCreate("aodv_hell", aodv_hello_thread, 0 , 512)) {//IMPROVE - Decide stack size
                                                                                               //Creating Thread failed, AODV will not operate
        ROUTING_DEBUG("\nE: (aodv_init) Create hello thread failed\n");
        return -1;
    }
    
    return 0;
}

/**
* Interface
 *
 * Provides the interface to the RFC layer.
 **/
routing_interface_t aodv_interface = {
    .get_address     = aodv_get_address,
    .get_next_hop    = aodv_get_next_hop,
    .discover        = aodv_discover,
    .send_msg        = aodv_recv_msg,
    .reg_msg_handler = aodv_reg_msg_handler,
};

#endif
