/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 * 
 * $Id: rfc.c,v 1.2 2007/06/11 12:05:41 yuecelm Exp $
 */

#ifdef __BTNODE3__

#include <sys/event.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <string.h>
#include <sys/atom.h>
#include <stdio.h>
#include <cc/rfc.h>


/////////////////////////////////////////////////////////////////////////////////////
// Defines
/////////////////////////////////////////////////////////////////////////////////////
#define FREE                          0 //Marks slot in buffer or queue as free.
#define FULL                          1 //Marks slot in buffer or queue as occupied.
#define RFC_ERROR_CCC_SEND_FAILED    -2 //Means that ccc could not send a packet for any reason.
#define RFC_TYPE_DATA                 1 //Type for RFC data packets.
#define RFC_TYPE_ACK                  2 //Type for RFC data acknowledgement packets.

/////////////////////////////////////////////////////////////////////////////////////
// Typedefs
/////////////////////////////////////////////////////////////////////////////////////
typedef struct rfc_ack_packet_st {
    u_char type;                                 //Type of the packet. MUST be RFC_TYPE_ACK.
    u_short originator;                          //Originator, who first sent the packet which must be acknowledged.
    u_short seqno;                               //Sequence number of the packet that must be acknowledged.
} __attribute__((__packed__)) rfc_ack_packet_t;

typedef struct send_status_st {
    u_char status;                               //Status of the packet that the entry belongs to.
    u_short src;                                 //Source of the packet that the entry belongs to.
    u_short seqno;                               //Sequence number of the packet that the entry belongs to.
    u_short retry;                               //Number of processing retries of the packet that the entry belongs to.
    long nextshd;                                //Time, from when on the packet that the entry belongs to can be scheduled.
} __attribute__((__packed__)) send_status_t;

typedef struct recv_status_st {
    u_short src;                                 //Source of the packet that the entry belongs to.
    u_short seqno;                               //Sequence number of the packet that the entry belongs to.
    long lifetime;                               //How long the entry for this packet is valid.
} __attribute__((__packed__)) recv_status_t;

typedef struct recv_queue_st {
    u_char status;                               //Status of the received packet.
    ccc_packet_t *pkt;                           //Pointer to the received packet.
} __attribute__((__packed__)) recv_queue_t;

typedef struct packet_handler_st {
    u_char port;                                 //Port to which the application registering the packet handler listens.
    rfc_packet_handler_t *handler;               //Pointer to the packet handler of the application.
} packet_handler_t;


/////////////////////////////////////////////////////////////////////////////////////
// Global Variables
/////////////////////////////////////////////////////////////////////////////////////

//Routing interface
static routing_interface_t *routing = NULL;

//Node Properties
u_short rfc_address;                     //Address of the node
u_short rfc_seqno;                       //Sequence no of the next packet

// Mutexes
MUTEX mutex_rfc_send_queue;
MUTEX mutex_rfc_send_status_queue;
MUTEX mutex_rfc_recv_queue_count;
MUTEX mutex_rfc_packet_handler;

// Handles
HANDLE event_rfc_packet_pending;        //Signals if a packet changed state in send queue and needs action
HANDLE event_packet_recvd;              //Signals if a rfc data packet is received.


// Queues & queue control 
recv_queue_t         rfc_recv_queue[RFC_RECV_QUEUE_SIZE];                // Buffer for received packets
u_short              rfc_recv_queue_count;                               // Number of packets in buffer
recv_status_t        rfc_recv_status_buffer[RFC_RECV_STATUS_BUFFER_SIZE];// Buffers already seen data packets

rfc_packet_t         *rfc_send_queue[RFC_SEND_QUEUE_SIZE];               // Buffer for packets waiting to be sent
u_short              rfc_send_queue_count;                               // Number of packets in buffer
send_status_t        rfc_send_status_queue[RFC_SEND_QUEUE_SIZE];         // Buffer for status messages

u_short              rfc_packet_handler_count;                           // Number of registered packet handlers
packet_handler_t     rfc_packet_handlers[RFC_PACKET_HANDLER_QUEUE_SIZE]; // how many ports can an node listen to?


// LED control
static u_char        rfc_use_led = 0;  //Controls wheter LED signaling is used or not

// Ack control
static u_char        rfc_use_ack = 0;  //Controls wheter DATA-ACKs are used or not

/////////////////////////////////////////////////////////////////////////////////////
// Local function predefinitions
/////////////////////////////////////////////////////////////////////////////////////
//Send functionallity
static inline void rfc_pkt_serialize(char *to, rfc_packet_t *from);
static int rfc_send_next_hop(u_short next_hop_addr, rfc_packet_t *pkt);
static int rfc_send_ack(u_short ccc_src, u_short originator, u_short seqno);

//Receive funtionallity
//static inline void rfc_pkt_deserialize(rfc_packet_t *to, char *from);

//Packet Management
static inline void rfc_pkt_serialize(char *to, rfc_packet_t *from);
static inline void rfc_pkt_deserialize(rfc_packet_t *to, char *from);
static inline void rfc_ack_pkt_serialize(char *to, rfc_ack_packet_t *from);
static inline void rfc_ack_pkt_deserialize(rfc_ack_packet_t *to, char *from);

//Application Level Packet Handling
int rfc_register_packet_handler(u_char port, rfc_packet_handler_t *handler);
void rfc_deregister_packet_handler(u_char port);
rfc_packet_handler_t *rfc_get_packet_handler(u_char port);

//Message Handling
static void routing_message_handler(u_short dst, u_short msg);

//Queue Handling
int rfc_recv_status_buffer_add(u_short seqno, u_short src);
int rfc_recv_status_buffer_get_handle(u_short seqno, u_short src);
void rfc_recv_status_buffer_touch(int handle);
int rfc_send_status_queue_add(u_short seqno, u_char status);
int rfc_send_status_queue_add_src(u_short src, u_short seqno, u_char status);
int rfc_send_status_queue_get_handle(u_short src, u_short seqno);
static inline void rfc_send_status_queue_set_status(int handle, u_char status);
static inline void rfc_send_status_queue_set_nextshd(int handle, long nextshd);
static inline int rfc_send_status_queue_inc_retry(int handle);
static inline int rfc_send_status_queue_update(int handle, u_char status, long nextshd); 
void rfc_send_status_queue_clean(void);

static inline void rfc_send_queue_del(int handle); 



/////////////////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////////////////

// calculates a waiting time with [ 1 .. 1 << tries  ] * slotduration
u_short rfc_beb(u_short slotduration, u_char next_try){
    if (next_try > 15) next_try = 15;
    u_short t_random = (rand() & ((1 << next_try) - 1) ) * slotduration;
    return slotduration + t_random;
}

/************************************************************************************
* Send Logic
************************************************************************************/

/**
* Puts packets in the send buffer.
 *
 * This function copies the packet that should be sent to the send buffer. It then
 * enters and updates all fields necessary. Note, that since this is done with the copy,
 * The original packet is not changed. 
 *
 * Once the packet is copied into the send buffer and all necessary fields are updated,
 * this function signals the send thread that a packet is waiting. 
 *
 * @param dst [in] destination where the packet should be delivered
 * @param port [in] port where the packet should be delivered
 * @param *pkt [in] pointer to the packet that should be sent
 * @return sequence no. on success, -1 otherwise
 */
u_short rfc_send(u_short dst, u_char port, rfc_packet_t *pkt)
{
    u_short i;
    
    
    //Aquire Mutexes
    NutMutexLock(&mutex_rfc_send_queue);
    
    if(rfc_send_queue_count < RFC_SEND_QUEUE_SIZE) { 
        
        //Enough space in queue, check if there is space in status queue. If there is, prepare an entry
        if(rfc_send_status_queue_add(rfc_seqno,RFC_STATUS_PENDING) != 0) {
            //Not enough space in status queue
            //Drop finalized packets, even if client looses track of them. (He didn't bother
            //much anyways).
            rfc_send_status_queue_clean();
            if(rfc_send_status_queue_add(rfc_seqno,RFC_STATUS_PENDING) != 0) {
                //No more space in queue, return with error
                ROUTING_DEBUG("\nE: (rfc_send) RFC send status queue full!\n");
                
                //Release mutex and return error
                NutMutexUnlock(&mutex_rfc_send_queue);
                return RFC_ERROR_SEND_QUEUE_FULL;
            }
        }
        
        //put a copy of the packet to next free postion
        for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++)
            if(rfc_send_queue[i] == NULL)
                break;
        
#ifdef ROUTING_DEBUGMODE
        if(i == RFC_SEND_QUEUE_SIZE) {
            //Error, should not happen
            ROUTING_DEBUG("\n!!! E: (rfc_send) Fatal error in send, smoke on the horizon (bo or mem leak)!\n");
        }
#endif //ROUTING_DEBUGMODE
        
        //Copy packet in queue at position i
        if( (rfc_send_queue[i] = new_rfc_packet(pkt->length)) == NULL) {
            //Memory allocation failed
            ROUTING_DEBUG("\nE: (rfc_send) RFC send could not allocate space for RFC data packet!\n");
            NutMutexUnlock(&mutex_rfc_send_queue);
            return -1;
        } 
        memcpy(rfc_send_queue[i], pkt, sizeof(rfc_packet_t) + pkt->length);
        
        //Update local data and register Packet Data   
        rfc_send_queue[i]->type = RFC_TYPE_DATA;         //This is a data packet
        rfc_send_queue[i]->seqno = rfc_seqno;            //Register the sequence number
        rfc_seqno = (rfc_seqno+1)&0x7FFF;                //Adjust seqno and check for ERRORMESSAGE
        if(rfc_use_ack)
            rfc_send_queue[i]->flags = RFC_FLAG_NEXT_HOP_ACK;
        else
            rfc_send_queue[i]->flags = 0;
        rfc_send_queue[i]->src = routing->get_address(); //Register the source address
        rfc_send_queue[i]->dst = dst;                    //Register the destination address
        rfc_send_queue[i]->port = port;                  //Register the port
        
        //Adjust queue counter
        rfc_send_queue_count++;
        
        //Release mutex
        NutMutexUnlock(&mutex_rfc_send_queue);
        
        //Set LED that data is pending
        if(rfc_use_led)
            btn_led_set(RFC_LED_DATA);
        
        
        //Post Event
        NutEventPost(&event_rfc_packet_pending);
        
        return rfc_seqno-1;
    } else {
        //Release mutex
        NutMutexUnlock(&mutex_rfc_send_queue);
        
        //Queue full, return error
        return RFC_ERROR_SEND_QUEUE_FULL;
    } 
}


/**
* Puts packets in the send buffer.
 *
 * This function copies the packet that must be forwarded to the send buffer. It then
 * enters and updates all fields necessary.
 *
 * Once the packet is copied into the send buffer and all necessary fields are updated,
 * this function signals the send thread that a packet is waiting. 
 *
 * @param [in] destination where the packet should be delivered
 * @param [in] port where the packet should be delivered
 * @param [in] pointer to the packet that should be sent
 * @return sequence no. on success, -1 otherwise
 */
u_short rfc_forward(rfc_packet_t *pkt)
{
    u_short i;
    
    //Create Packet and put it in the send queue. Then signal that
    //the state of a packet in the send queue changed. 
    
    NutMutexLock(&mutex_rfc_send_queue);
    
    //ROUTING_DEBUG("\tLooking up space in queue (forward)...(%u)",rfc_send_queue_count);
    
    if(rfc_send_queue_count < RFC_SEND_QUEUE_SIZE) {
        
        //Enough space in queue, check if there is space in status queue. If there is, prepare an entry
        if(rfc_send_status_queue_add_src(pkt->src, pkt->seqno,RFC_STATUS_PENDING) != 0) {
            //Not enough space in status queue
            //Drop finalized packets, even if client looses track of them. (He didn't bother
            //much anyways).
            rfc_send_status_queue_clean();
            if(rfc_send_status_queue_add_src(pkt->src, pkt->seqno,RFC_STATUS_PENDING) != 0) {
                //No more space in queue, return with error
                ROUTING_DEBUG("\nE: (rfc_forward) RFC send status queue full!\n");
                
                //Release mutex and return with error
                NutMutexUnlock(&mutex_rfc_send_queue);
                return RFC_ERROR_SEND_QUEUE_FULL;
            }
        }
        
        //Put a copy of the packet to next free postion
        for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++)
            if(rfc_send_queue[i] == NULL)
                break;
        
#ifdef ROUTING_DEBUGMODE
        if(i == RFC_SEND_QUEUE_SIZE) {
            //Error, should not happen
            ROUTING_DEBUG("\nA: (rfc_forward) Fatal error in forward, memory leak or bo guaranteed!\n");
        }
#endif //ROUTING_DEBUGMODE
        
        if( (rfc_send_queue[i] = new_rfc_packet(pkt->length)) == NULL) {
            ROUTING_DEBUG("\nE: (rfc_send) RFC send could not allocate space for RFC data packet!\n");
            NutMutexUnlock(&mutex_rfc_send_queue);
            return -1;
        } 
        memcpy(rfc_send_queue[i], pkt, sizeof(rfc_packet_t) + pkt->length);
        
        //Adjust flags in packet
        if(rfc_use_ack)
            rfc_send_queue[i]->flags = RFC_FLAG_NEXT_HOP_ACK;
        else
            rfc_send_queue[i]->flags = 0;
        
        //Adjust queue counters
        rfc_send_queue_count++;
        
        //Release mutex
        NutMutexUnlock(&mutex_rfc_send_queue);
        
        //Set LED that data is pending
        if(rfc_use_led)
            btn_led_set(RFC_LED_DATA);
        
        //Post Event
        NutEventPost(&event_rfc_packet_pending);
        
        return 0;
    } else {
        
        //Release mutex
        NutMutexUnlock(&mutex_rfc_send_queue);
        
        //Queue full, return error
        return RFC_ERROR_SEND_QUEUE_FULL;
    } 
}



/** 
* Control the rfc send queue
*
*/
THREAD(rfc_send_thread,arg)
{
    int i, retry_count = 0;
    u_short next_hop = 0;
    long t_next_shd, t_wait;
    u_long t_now, t_my_delta;
    int done;
    
    for(;;) {
    	
	//printf( "THREAD rfc_send_thread\n" );    	
        
        //Initialize the wait time to INFINITE
        t_wait = NUT_WAIT_INFINITE;
        t_next_shd = -1;
        
        //Lock the send queue and process every packet in it
        NutMutexLock(&mutex_rfc_send_queue);
 
        for(i = 0, done = 0; !done && rfc_send_queue_count > 0 && i < RFC_SEND_QUEUE_SIZE; i++) {

            // do we have packet?
            if(rfc_send_queue[i] == NULL) {
                continue;
            }
            
            //Get handle to map from send queue to status queue
            int status_handle = rfc_send_status_queue_get_handle(rfc_send_queue[i]->src, rfc_send_queue[i]->seqno);
            
            //If status handle was unavailable, an error occured. Drop packet
            if(status_handle == -1) { 
                ROUTING_DEBUG("\nW: (rfc_send_thread) \n\tPacket without valid status handle, removing from queue).\n");
                rfc_send_queue_del(i);
                // HACK: tweak index i, as next entry will be on current index
                i--;
                continue;
            }
            
            // NEW: either there is an action scheduled or it will be executed, we're done anyway
            done = 1;
            
            // Check if this packet must already be processed, if not skip to next packet
            if(rfc_send_status_queue[status_handle].nextshd > NutGetMillis()) {
                //Update t_next_shd
                if(rfc_send_queue[i] != NULL && 
                   rfc_send_status_queue[status_handle].nextshd != NUT_WAIT_INFINITE && 
                   (t_next_shd == -1 || rfc_send_status_queue[status_handle].nextshd < t_next_shd)) {
                    t_next_shd = rfc_send_status_queue[status_handle].nextshd;	
                }	  
                continue;
            }
            
            switch(rfc_send_status_queue[status_handle].status) {
                /* CASE 0: Packet is waiting for ack */
                case RFC_STATUS_WAIT_NEXT_HOP_ACK:	
                    //Fall through to RFC_STATUS_READY
                    
                    /* CASE 1: Route (was) available, try to send packet out */
                case RFC_STATUS_READY:	  
                    //Check whether route still exists
                    if( (next_hop = routing->get_next_hop(&(rfc_send_queue[i]->dst))) ) {
                        
                        //Next hop known - get rid of that packet	  
                        if(rfc_send_next_hop(next_hop, rfc_send_queue[i]) != RFC_NO_ERROR) {
                            //Sending failed, so we increment the retry counter and update the next schedule.
                            retry_count = rfc_send_status_queue_inc_retry(status_handle);
                            rfc_send_status_queue_set_nextshd(status_handle, 
                                                    NutGetMillis() + rfc_beb(RFC_SEND_QUEUE_RETRY_TIMEOUT, retry_count));
                        } else {
                            //Packet successfully sent
                            //ACK behaviour
                            if(rfc_send_queue[i]->flags & RFC_FLAG_NEXT_HOP_ACK) {
                                retry_count = rfc_send_status_queue_inc_retry(status_handle);
                                t_now = NutGetMillis();
                                t_my_delta = rfc_beb(RFC_SEND_QUEUE_RETRY_TIMEOUT, retry_count);
                                rfc_send_status_queue_set_nextshd(status_handle, t_now + t_my_delta);	
                                rfc_send_status_queue_set_status(status_handle, RFC_STATUS_WAIT_NEXT_HOP_ACK);
                            } else {		
                                //Mark packet status as delivered successfully to next hop and delete it from the queue
                                rfc_send_status_queue_set_status(status_handle, RFC_STATUS_SENT_TO_NEXT_HOP);
                                rfc_send_queue_del(i);
                                // HACK: tweak index i, as next entry will be on current index
                                i--;
                                
                                //Packet is gone, clear data LED
                                if(rfc_use_led)
                                    btn_led_clear(RFC_LED_DATA);
                            }
                        }
                    } else {
                        //Next hop not known - this might happen if a route reached its end of life
                        //If this is the originator, mark packet as status pending and retry. 
                        //Either the route can not be established again and it will be deleted by message handler, 
                        //or it will be delivered.
                        //If this is an intermeadiate node, inform routing protocol and wait for instructions.
                        if(rfc_send_queue[i]->src == routing->get_address()) {
                            //I am the originator, simply try a new discovery
                            retry_count = rfc_send_status_queue_update(status_handle, RFC_STATUS_PENDING, 
                                NutGetMillis() + rfc_beb(RFC_SEND_QUEUE_RETRY_TIMEOUT, rfc_send_status_queue[status_handle].retry + 1));
#ifdef ROUTING_INFOMODE
                            ROUTING_DEBUG("\nI: (rfc_send_thread) Route timed out before packet got sent. (retry: %u in %ldms)\n",
                                          retry_count,rfc_send_status_queue[status_handle].nextshd - NutGetMillis());
#endif //ROUTING_INFOMODE
                        } else {
                            //I am an intermediate node. Inform routing protocol.
                            routing->send_msg(BROKEN_ROUTE, rfc_send_queue[i]->dst);
                            
                            //Mark packet as waiting for a message from routing protocol
                            rfc_send_status_queue_set_status(status_handle, RFC_STATUS_WAITING);
                            rfc_send_status_queue_set_nextshd(status_handle, NUT_WAIT_INFINITE); 
                        }
                    }
                    break;
                    
                    
                    /* CASE 2: Either this is a new packet or a packet that could not be processed yet. */	  
                case RFC_STATUS_PENDING:
                    
                    // Check whether route exists
                    // There are three case. Case one is the special case that the node sent the data to itself. Since
                    // BTNodes do not have a loopback interface, we make a shortcut here and directly dispatch the 
                    // packet. Case two and three are that a route to the destination is, respectively is not available.
                    if(rfc_send_queue[i]->dst == routing->get_address()) {
                        
                        //1. I am the destination myself, directly dispatch packet.
                        rfc_packet_handler_t *tmp_handler = rfc_get_packet_handler(rfc_send_queue[i]->port);
                        if(tmp_handler != NULL) {
                            //Dispatch packet and mark it as delivered.
                            ROUTING_DEBUG("\n<<< DATA from %u to %u seqno %u\n", 
                                          rfc_send_queue[i]->dst, routing->get_address(), rfc_send_queue[i]->seqno);
                            tmp_handler->packet_handler(rfc_send_queue[i]);
                            rfc_send_status_queue_set_status(status_handle, RFC_STATUS_SENT);
                        } else {
                            //Packet could not be delivered, do error handling and mark packet as rejected.
                            ROUTING_DEBUG("\nW: (rfc_send_thread) No handler for pkt registered.\n");
                            rfc_send_status_queue_set_status(status_handle, RFC_STATUS_REJECTED);
                        }
                        
                        //Delete Packet from queue
                        rfc_send_queue_del(i);
                        // HACK: tweak index i, as next entry will be on current index
                        i--;
                        
                        if(rfc_use_led)
                            btn_led_clear(RFC_LED_DATA);
                        
                    } else if( (next_hop = routing->get_next_hop(&(rfc_send_queue[i]->dst))) ) {
                        
                        //2. Next hop known - mark packet ready for sending immediately.
                        rfc_send_status_queue_set_status(status_handle, RFC_STATUS_READY);
                        rfc_send_status_queue_set_nextshd(status_handle, NutGetMillis());
                        
                    } else {
                        
                        //3. Next hop not known
                        
                        //Start discovery
                        routing->discover(&(rfc_send_queue[i]->dst));
                        
                        //Mark packet as waiting for a message from routing protocol
                        rfc_send_status_queue_set_status(status_handle, RFC_STATUS_WAITING);
                        rfc_send_status_queue_set_nextshd(status_handle, NUT_WAIT_INFINITE); 
                        
                    }
                    break;
            }
            
            //If the packet still exists, check wheter it must be dropped or not
            if(rfc_send_queue[i] != NULL &&
               rfc_send_status_queue[status_handle].retry == RFC_SEND_RETRY) {
                rfc_send_status_queue_set_status(status_handle, RFC_STATUS_FAILED);
                rfc_send_queue_del(i);
                // HACK: tweak index i, as next entry will be on current index
                i--;
                
                //Disble LED, packet has been deleted
                if(rfc_use_led)
                    btn_led_clear(RFC_LED_DATA);
                
                //If there was a next hop known for this packet, and it reached the max number
                //of retries, most likely this link is broken. Inform the routing protocoll about that.
                if(next_hop != 0)
                    routing->send_msg(BROKEN_LINK, next_hop);
            } 
            
            //If packet still exists, calculate t_wait and update it, if this packet must be
            //scheduled before any previously processed packet.
            if(rfc_send_queue[i] != NULL && 
               rfc_send_status_queue[status_handle].nextshd != NUT_WAIT_INFINITE && 
               (t_next_shd == -1 || rfc_send_status_queue[status_handle].nextshd < t_next_shd)) {
                t_next_shd = rfc_send_status_queue[status_handle].nextshd;	
            }
        }
        
        //Unlock Mutex
        NutMutexUnlock(&mutex_rfc_send_queue);
        
        //If the next schedule was updated this turn, calculate the wait time. Otherwise wait INFINITE
        if(t_next_shd != -1) {
            t_wait = t_next_shd - NutGetMillis();
            //Wait for next turnaround if t_wait is positive
            if(t_wait > 0) {
                NutEventWait(&event_rfc_packet_pending, t_wait);
            }
        } else {
            NutEventWait(&event_rfc_packet_pending, NUT_WAIT_INFINITE);
        }
    }
}


/**
* Send packet to next hop
 *
 */
static int rfc_send_next_hop(u_short next_hop_addr, rfc_packet_t *pkt)
{
    ccc_packet_t *ccc_pkt = NULL;
    
    if( (ccc_pkt = new_ccc_packet(sizeof(rfc_packet_t) + pkt->length)) == NULL) {
        //Allocation failed
        ROUTING_DEBUG("\nE: (rfc_send_next_hop) Packet allocation failed.\n");
        
        return RFC_ERROR_CCC_SEND_FAILED;
    }
    
    ccc_pkt->length = sizeof(rfc_packet_t) + pkt->length;
    
    //Serialize RFC Packet
    rfc_pkt_serialize((char *) ccc_pkt->data, pkt);
    
    //Send packet out
    if(ccc_send(next_hop_addr, RFC_TYPE, ccc_pkt)) {
        //Send failed
        free(ccc_pkt);
        ccc_pkt = NULL;
        
        return RFC_ERROR_CCC_SEND_FAILED;
    } else {
        ROUTING_DEBUG("\n>>> DATA from %u to %u seqno %u\n", pkt->src, pkt->dst, pkt->seqno);
    }
    
    //Free ccc_packet
    free(ccc_pkt);
    ccc_pkt = NULL;
    
    return RFC_NO_ERROR;
}


static int rfc_send_ack(u_short ccc_src, u_short originator, u_short seqno)
{
    rfc_ack_packet_t ack;
    ccc_packet_t *ccc_pkt;
    
    //Fill in header information
    ack.type = RFC_TYPE_ACK;
    ack.originator = originator;
    ack.seqno = seqno;
    
    //Create ccc packet
    if( (ccc_pkt = new_ccc_packet(sizeof(rfc_ack_packet_t))) == NULL) {
        //Allocation failed
        ROUTING_DEBUG("\nE: (rfc_send_ack) Space allocation for DATA-ACK failed\n");
        return -1;
    }
    
    ccc_pkt->length = sizeof(rfc_ack_packet_t);
    
    //Serialize RFC Packet
    rfc_ack_pkt_serialize((char *) ccc_pkt->data, &ack);
    
    //Send packet out
    if(ccc_send(ccc_src, RFC_TYPE, ccc_pkt)) {
        //Send failed
        free(ccc_pkt);
        ccc_pkt = NULL;
        return -1;
    } else { 
        ROUTING_DEBUG("\n>>> DATA-ACK to %u for seqno\n", ccc_src, seqno);
    }
    
    //Free ccc_packet
    free(ccc_pkt);
    ccc_pkt = NULL;
    
    return 0;
}


//////////////////////////////////////////////////////////////////////////
// Receive Logic
/////////////////////////////////////////////////////////////////////////

/**
* Thread that processes received RFC packets
 **/
THREAD(rfc_recv_thread,arg)
{
    int queue_entry,i;
    rfc_packet_t *rfc_pkt;
    rfc_ack_packet_t *rfc_ack_pkt;
    int recv_status_handle = -1;
    //u_short next_hop;
    
    for(;;) {
    	
    	//printf( "THREAD rfc_recv_thread\n" );
        
        // Process next entry in queue if there is any.
        NutMutexLock(&mutex_rfc_recv_queue_count);
        if(rfc_recv_queue_count > 0) { 
            
            for(queue_entry = 0; rfc_recv_queue_count > 0 && queue_entry < RFC_SEND_QUEUE_SIZE; queue_entry++) {
                
                if(rfc_recv_queue[queue_entry].status != FREE) {
                    //Valid Entry, process	  
                    switch(rfc_recv_queue[queue_entry].pkt->data[0]) {
                        case RFC_TYPE_DATA:  
                            //Get Packet
                            if( (rfc_pkt = new_rfc_packet(rfc_recv_queue[queue_entry].pkt->length - sizeof(rfc_packet_t))) != NULL) {
                                rfc_pkt_deserialize(rfc_pkt, (char *) rfc_recv_queue[queue_entry].pkt->data);
                                
                                //ACK Packet
                                if(rfc_pkt->flags & RFC_FLAG_NEXT_HOP_ACK) {
                                    rfc_send_ack(rfc_recv_queue[queue_entry].pkt->src, rfc_pkt->src, rfc_pkt->seqno);
                                }
                                
                                ROUTING_DEBUG("<<< DATA from %u to %u with seqno %u\n", rfc_pkt->src, rfc_pkt->dst, rfc_pkt->seqno);
                                
                                //If packet not yet seen: process, else drop
                                if( (recv_status_handle = rfc_recv_status_buffer_get_handle(rfc_pkt->seqno, rfc_pkt->src)) == -1) { 
                                    
                                    //Set the data LED
                                    if(rfc_use_led)
                                        btn_led_set(RFC_LED_DATA);
                                    
                                    //Add to status buffer
                                    if(rfc_recv_status_buffer_add(rfc_pkt->seqno, rfc_pkt->src) == 0) {
                                        //Packet could be added
                                        
                                        //Decide what to do
                                        if(rfc_pkt->dst == routing->get_address()) {
                                            //I am the destination, deliver to listener
                                            // dispatch this packet to the application
                                            rfc_packet_handler_t *tmp_handler = rfc_get_packet_handler(rfc_pkt->port);
                                            if(tmp_handler != NULL) {
                                                //Dispatch
                                                tmp_handler->packet_handler(rfc_pkt);
                                                
                                                //Clear data led
                                                if(rfc_use_led)
                                                    btn_led_clear(RFC_LED_DATA);
                                            } else {
                                                ROUTING_DEBUG("\nW: (rfc_recv_thread) No handler for pkt registered.\n");
                                            }
                                            
                                        } else {
                                            //Forward packet to next hop
                                            rfc_forward(rfc_pkt);
                                        }
                                    } else {
                                        //Failed adding to status buffer, drop
                                        ROUTING_DEBUG("\nW: (rfc_recv_thread) Recv status buffer full, dropping packet.\n");
                                    }
                                } else {
                                    //Refresh entry on status buffer
                                    rfc_recv_status_buffer_touch(recv_status_handle);
#ifdef ROUTING_INFOMODE
                                    ROUTING_DEBUG("\nI: (rfc_recv_thread) Got data packet twice, dropping.\n");
#endif //ROUTING_INFOMODE
                                }
                                
                                //Free packet
                                free(rfc_pkt);
                                rfc_pkt = NULL;
                            } else {
                                ROUTING_DEBUG("\nE: (rfc_recv_thread) Could not allocate space for RFC data packet.\n");
                            }
                            break;
                            
                        case RFC_TYPE_ACK:
                            if( (rfc_ack_pkt = malloc(sizeof(rfc_ack_packet_t))) != NULL) {
                                rfc_ack_pkt_deserialize(rfc_ack_pkt, (char *) rfc_recv_queue[queue_entry].pkt->data);
                                
                                ROUTING_DEBUG("\n<<< DATA-ACK from %u for seqno %u\n",
                                              rfc_recv_queue[queue_entry].pkt->src, rfc_ack_pkt->seqno);
                                
                                //Remove entry from send queue and set status
                                NutMutexLock(&mutex_rfc_send_queue);
                                for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
                                    if(rfc_send_queue[i] != NULL &&
                                       rfc_send_queue[i]->src == rfc_ack_pkt->originator && 
                                       rfc_send_queue[i]->seqno == rfc_ack_pkt->seqno) { 
                                        
                                        //Delete packet from queue and mark it as delivered to next hop
                                        int status_handle = 
                                        rfc_send_status_queue_get_handle(rfc_send_queue[i]->src, 
                                                                         rfc_send_queue[i]->seqno);
                                        if(status_handle != -1) {
                                            rfc_send_queue_del(i);
                                            // HACK: tweak index i, as next entry will be on current index
                                            i--;
                                            
                                            //Clear data led
                                            if(rfc_use_led)
                                                btn_led_clear(RFC_LED_DATA);
                                            rfc_send_status_queue_set_status(status_handle, RFC_STATUS_SENT_TO_NEXT_HOP);	
                                            break;
                                        }
#ifdef ROUTING_DEBUGMODE
                                        else {
                                            ROUTING_DEBUG("\nW: Could not find status handle that matches packet.\n");
                                            break;
                                        }
#endif //ROUTING_DEBUGMODE
                                    }
                                }
                                
#ifdef ROUTING_DEBUGMODE
                                if(i == RFC_SEND_QUEUE_SIZE) {
                                    ROUTING_DEBUG("\nW: ACK could not be matched to an entry in the send queue.\n");
                                }
#endif //ROUTING_DEBUGMODE
                                
                                NutMutexUnlock(&mutex_rfc_send_queue);

                                // Post Event
                                NutEventPost(&event_rfc_packet_pending);

                                //Free 
                                if(rfc_ack_pkt != NULL) {
                                    free(rfc_ack_pkt);
                                    rfc_ack_pkt = NULL;
                                }
#ifdef ROUTING_DEBUGMODE
                                else {
                                    ROUTING_DEBUG("--> E: RFC ACK is not NULL");
                                }
#endif //ROUTING_DEBUGMODE
                            } else {
                                ROUTING_DEBUG("\n!!! A: Packet allocation for RFC ACK failed in recv thread.\n");
                            }
                            break;
                    } //end switch
                    
                    //Remove packet from queue
#ifdef ROUTING_DEBUGMODE
                    if(rfc_recv_queue[queue_entry].pkt == NULL) {
                        ROUTING_DEBUG("\n!!! A: Havoc, freeing null pointer. Expect everything!\n");
                    }
#endif //ROUTING_DEBUGMODE
                    free(rfc_recv_queue[queue_entry].pkt);
                    rfc_recv_queue[queue_entry].pkt = NULL;
                    rfc_recv_queue_count--;
                    rfc_recv_queue[queue_entry].status = FREE;
                } // end if
            } // end for
        } // end if queue_count > 0
        
        NutMutexUnlock(&mutex_rfc_recv_queue_count);
        
        //Wait until there is something to send
        NutEventWait(&event_packet_recvd, NUT_WAIT_INFINITE);
    }
}


/**
* CCC handler for receiving RFC packets
 **/
static void rfc_receive_handler(ccc_packet_t *pkt)
{
    u_short i;
    
    // Check if there is enough space in the queue
    if(rfc_recv_queue_count == RFC_RECV_QUEUE_SIZE) {
#ifdef ROUTING_DEBUGMODE
        fflush(stdout); printf("\nW: (rfc_receive_handler) Queue full, packet ignored!\n"); fflush(stdout);
#endif //ROUTING_DEBUGMODE
        return;
    } else {
        //Search for free position
        for(i = 0; i < RFC_RECV_QUEUE_SIZE; i++) {
            if(rfc_recv_queue[i].pkt == NULL && rfc_recv_queue[i].status == FREE)
                break;
#ifdef ROUTING_DEBUGMODE
            else if(rfc_recv_queue[i].status == FREE && rfc_recv_queue[i].pkt != NULL) {
                //Shouldn't happen, really!
                fflush(stdout); printf("\n !!! A: (rfc_receive_handler) Memory leak possible...\n"); fflush(stdout);
            }
#endif //ROUTING_DEBUGMODE
        }
#ifdef ROUTING_DEBUGMODE
        if(i == RFC_RECV_QUEUE_SIZE) {
            //Shouldn't happen, really!
            fflush(stdout); printf("\n!!! A: (rfc_receive_handler) Counter desync...\n"); fflush(stdout);
        }
#endif //ROUTING_DEBUG
    }
    
    // Reserve space for packet
    if( (rfc_recv_queue[i].pkt = malloc(sizeof(ccc_packet_t) + pkt->length)) != 0 ) {
        //Success
        memcpy(rfc_recv_queue[i].pkt,pkt,(sizeof(ccc_packet_t) + pkt->length));
        rfc_recv_queue[i].status = FULL;
        
        //Increase number of packets in queue
        rfc_recv_queue_count++;
        
        //Signal that there is a new packet, just in case the receiver thread sleeps.
        NutEventPostAsync(&event_packet_recvd);
    } else { 
        // Space allocation failed, drop packet (==abort)
#ifdef ROUTING_DEBUGMODE
        fflush(stdout); printf("\nE: (rfc_receive_handler) Not enough space for packet!\n"); fflush(stdout);
#endif //ROUTING_DEBUGMODE
    }
}



/************************************************************************************
* Packet Management
************************************************************************************/

/** 
* Packet allocation
*/
rfc_packet_t *new_rfc_packet(int data_len)
{
    return (rfc_packet_t *)malloc(data_len + sizeof(rfc_packet_t));
}


/**
* Serialize RFC packet to char buffer 
 *
 * Used for sending the packet as payload in a ccc_packet
 */  
static inline void rfc_pkt_serialize(char *to, rfc_packet_t *from)
{
    u_short len;
    
    to[0] = from->type;
    to[1] = from->port;
    *((u_short*) &to[2]) = htons((u_short)from->seqno);
    *((u_short*) &to[4]) = htons((u_short)from->flags);
    *((u_short*) &to[6]) = htons((u_short)from->src);
    *((u_short*) &to[8]) = htons((u_short)from->dst);
    *((u_short*) &to[10]) = htons((u_short)from->length);
    
    for(len = 0; len < from->length; len++)
        to[len+12] = from->data[len];
}


/**
* Deserialize RFC packet from char buffer 
 *
 * Used for getting the packet from ccc_packet payload
 */  
static inline void rfc_pkt_deserialize(rfc_packet_t *to, char *from)
{
    u_short len;
    
    to->type = from[0];
    to->port = from[1];
    to->seqno = ntohs(*((u_short*) &from[2]));
    to->flags = ntohs(*((u_short*) &from[4]));
    to->src = ntohs(*((u_short*) &from[6]));
    to->dst = ntohs(*((u_short*) &from[8]));
    to->length = ntohs(*((u_short*) &from[10]));
    
    for(len = 0; len < to->length; len++)
        to->data[len] = from[len+12];
}


/**
* Serialize RFC ACK packet to char buffer 
 *
 * Used for sending the packet as payload in a ccc_packet
 */  
static inline void rfc_ack_pkt_serialize(char *to, rfc_ack_packet_t *from)
{
    to[0] = from->type;
    *((u_short*) &to[1]) = htons((u_short)from->originator);
    *((u_short*) &to[3]) = htons((u_short)from->seqno);
}

/**
* Deserialize RFC ACK packet from char buffer 
 *
 * Used for getting the packet from ccc_packet payload
 */  
static inline void rfc_ack_pkt_deserialize(rfc_ack_packet_t *to, char *from)
{
    to->type = from[0];
    to->originator = ntohs(*((u_short*) &from[1]));
    to->seqno = ntohs(*((u_short*) &from[3]));
}


/************************************************************************************
* Application Level Packet Handling
************************************************************************************/

int rfc_register_packet_handler(u_char port, rfc_packet_handler_t *handler)
{
    u_char i;
    if (port == 0 || rfc_packet_handler_count >= RFC_PACKET_HANDLER_QUEUE_SIZE) return -1;
    
    NutMutexLock(&mutex_rfc_packet_handler);
    for (i = 0; i < rfc_packet_handler_count; i++) {
        if (rfc_packet_handlers[i].port == port) {
            NutMutexUnlock(&mutex_rfc_packet_handler);
            return -1;        
        }
    }
    
    rfc_packet_handlers[i].port = port;
    if( (rfc_packet_handlers[i].handler = malloc(sizeof(rfc_packet_handler_t))) == NULL) {
        ROUTING_DEBUG("\nE: (rfc_register_packet_handler) Could not allocate memory for handler structure.\n");
        return -1;
    }
    rfc_packet_handlers[i].handler->packet_handler = handler->packet_handler;
    rfc_packet_handlers[i].handler->error_handler = handler->error_handler;
    
    ++rfc_packet_handler_count;
    NutMutexUnlock(&mutex_rfc_packet_handler);
    
    return 0;
}


void rfc_deregister_packet_handler(u_char port)
{
    u_char i, j;
    //Replace the specified handler with the last in the list 
    //and shorten the list by one 
    NutMutexLock(&mutex_rfc_packet_handler);
    for (i = 0; i < RFC_PACKET_HANDLER_QUEUE_SIZE && port != 0; i++) {
        if (rfc_packet_handlers[i].port == port) {
            j = --rfc_packet_handler_count;
            rfc_packet_handlers[i].port = rfc_packet_handlers[j].port;
            rfc_packet_handlers[i].handler = rfc_packet_handlers[j].handler;
            rfc_packet_handlers[j].port = 0;
            rfc_packet_handlers[j].handler = NULL;
            break;  
        }        
    }    
    NutMutexUnlock(&mutex_rfc_packet_handler);
}

rfc_packet_handler_t *rfc_get_packet_handler(u_char port)
{
    u_char i;
    NutMutexLock(&mutex_rfc_packet_handler);
    for (i = 0; i < RFC_PACKET_HANDLER_QUEUE_SIZE; i++) {
        if (rfc_packet_handlers[i].port == port) {
            NutMutexUnlock(&mutex_rfc_packet_handler);
            return rfc_packet_handlers[i].handler;
        }
    }
    NutMutexUnlock(&mutex_rfc_packet_handler);
    return NULL;
}

/**
* Get the maximal transfer unit
 *
 * SHOULD be used to decide maximum size of a packet that can be sent.
 *
 * @return Maximal payload size of a data packet.
 */
u_short rfc_get_mtu(void)
{
    return (ccc_get_mtu() - sizeof(rfc_packet_t));
}


/************************************************************************************
* Message Handling
************************************************************************************/

static void routing_message_handler(u_short dst, u_short msg)
{
    int i;
    
    switch(msg) {
        case DISCOVERY_STARTED:
            //IMPROVE - Set any packet in queue waiting for destination to state wait. This is not
            //necessary, the protocol will work flawlessly without this, but this would speed
            //up processing.
#ifdef ROUTING_INFOMODE
            ROUTING_DEBUG("\nI: (routing_message_handler) Discovery for %u started.\n", dst);
#endif //ROUTING_INFOMODE
            break;
            
        case DESTINATION_UNREACHABLE:
            //Set status for pending packets
            NutMutexLock(&mutex_rfc_send_queue);
            
            for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
                if(rfc_send_queue[i]->dst == dst) {
                    //Notify Application
                    rfc_packet_handler_t *tmp_handler = rfc_get_packet_handler(rfc_send_queue[i]->port);
                    if(tmp_handler != NULL) {
                        //Dispatch
                        tmp_handler->error_handler(DESTINATION_UNREACHABLE, rfc_send_queue[i]->seqno);
                    } else {
                        ROUTING_DEBUG("No handler for pkt registered.\n");
                    }
                    
                    //Mark as failed in status queue
                    int status_handle = rfc_send_status_queue_get_handle(rfc_send_queue[i]->src, rfc_send_queue[i]->seqno);
                    
                    NutMutexLock(&mutex_rfc_send_status_queue);
                    rfc_send_status_queue[status_handle].status = RFC_DESTINATION_UNREACHABLE;
                    NutMutexUnlock(&mutex_rfc_send_status_queue);
                    
                    //Delete pending packet
                    rfc_send_queue_del(i);
                    // HACK: tweak index i, as next entry will be on current index
                    i--;

                    //Clear LED
                    if(rfc_use_led)
                        btn_led_clear(RFC_LED_DATA);
                    
                }
            }
                
            NutMutexUnlock(&mutex_rfc_send_queue);

            // other packets might be ready
            NutEventPost(&event_rfc_packet_pending);
            
            break;
            
        case DISCOVERY_FAILED:
            //Set status for pending packets
            NutMutexLock(&mutex_rfc_send_queue);
            
            for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
                if(rfc_send_queue[i]->dst == dst) {
                    //Notify Application
                    rfc_packet_handler_t *tmp_handler = rfc_get_packet_handler(rfc_send_queue[i]->port);
                    if(tmp_handler != NULL) {
                        //Dispatch
                        tmp_handler->error_handler(DESTINATION_UNREACHABLE, rfc_send_queue[i]->seqno);
                    } else {
                        ROUTING_DEBUG("No handler for pkt registered.\n");
                    }
                    
                    //Mark as failed in status queue
                    int status_handle = 
                        rfc_send_status_queue_get_handle(rfc_send_queue[i]->src, 
                                                         rfc_send_queue[i]->seqno);
                    
                    NutMutexLock(&mutex_rfc_send_status_queue);
                    rfc_send_status_queue[status_handle].status = RFC_STATUS_FAILED;
                    NutMutexUnlock(&mutex_rfc_send_status_queue);
                    
                    //Delete pending packet
                    rfc_send_queue_del(i);
                    // HACK: tweak index i, as next entry will be on current index
                    i--;
                    
                    //Clear LED
                    if(rfc_use_led)
                        btn_led_clear(RFC_LED_DATA);
                }
            }
                
            NutMutexUnlock(&mutex_rfc_send_queue);
            
            // other packets might be ready
            NutEventPost(&event_rfc_packet_pending);
            
            break;
            
        case ROUTE_AVAILABLE:
#ifdef ROUTING_INFOMODE
            ROUTING_DEBUG("\nI: (routing_message_handler) Route for %u available.\n", dst);
#endif //ROUTING_INFOMODE
            
            //Set Packets READY so they will get sent by thread    
            NutMutexLock(&mutex_rfc_send_queue);
            
            for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
                if(rfc_send_queue[i]->dst == dst) {	
                    //Mark as ready in status queue
                    int status_handle = rfc_send_status_queue_get_handle(rfc_send_queue[i]->src, rfc_send_queue[i]->seqno);
                    
                    NutMutexLock(&mutex_rfc_send_status_queue);
                    rfc_send_status_queue[status_handle].status = RFC_STATUS_READY;
                    rfc_send_status_queue[status_handle].nextshd = NutGetMillis();
                    NutMutexUnlock(&mutex_rfc_send_status_queue);
                }
            }
                
            NutMutexUnlock(&mutex_rfc_send_queue);
            
            // other packets might be ready
            NutEventPost(&event_rfc_packet_pending);
            
            break;
    }
}

/************************************************************************************
* Queue Handling
************************************************************************************/

/**
* Add a rfc packet to the buffer.
 */
int rfc_recv_status_buffer_add(u_short seqno, u_short src)
{
    int i = 0;
    
    //Search free slot
    for(i = 0; i < RFC_RECV_STATUS_BUFFER_SIZE; i++) {
        if(rfc_recv_status_buffer[i].lifetime < NutGetMillis()) {
            rfc_recv_status_buffer[i].seqno = seqno;
            rfc_recv_status_buffer[i].src = src;
            rfc_recv_status_buffer[i].lifetime = NutGetMillis() + RFC_RECV_STATUS_BUFFER_TIMEOUT;
            break;
        }
    }
    
    if( i == RFC_RECV_STATUS_BUFFER_SIZE )
        return -1;
    else
        return 0;
}

int rfc_recv_status_buffer_get_handle(u_short seqno, u_short src)
{
    int i = 0;
    
    //Search entry
    for(i = 0; i < RFC_RECV_STATUS_BUFFER_SIZE; i++) {
        if(rfc_recv_status_buffer[i].lifetime > NutGetMillis() &&
           rfc_recv_status_buffer[i].seqno == seqno &&
           rfc_recv_status_buffer[i].src == src) {
            break;
        }
    }
    
    if( i == RFC_RECV_STATUS_BUFFER_SIZE )
        return -1;
    else
        return i;
}

void rfc_recv_status_buffer_touch(int handle)
{
    rfc_recv_status_buffer[handle].lifetime = NutGetMillis() + RFC_RECV_STATUS_BUFFER_TIMEOUT;
}

/**
* Add a new (local) entry into the send status queue
 */
int rfc_send_status_queue_add(u_short seqno, u_char status)
{
    return rfc_send_status_queue_add_src(routing->get_address(), seqno, status);
}

/**
* Add a new entry into the send status queue (might be local or remote!)
 */
int rfc_send_status_queue_add_src(u_short src, u_short seqno, u_char status)
{
    u_char i;
    
    NutMutexLock(&mutex_rfc_send_status_queue);
    for(i=0; i < RFC_SEND_QUEUE_SIZE; i++) {
        if(rfc_send_status_queue[i].status == RFC_STATUS_UNDEF)
            //Take this free slot
            break;
    }
    
    //Check if free entry was available
    if(i < RFC_SEND_QUEUE_SIZE) {
        //Enough space in queue, accept
        rfc_send_status_queue[i].src = src;
        rfc_send_status_queue[i].seqno = seqno;
        rfc_send_status_queue[i].status = status;
        rfc_send_status_queue[i].retry = 0;
        rfc_send_status_queue[i].nextshd = NutGetMillis(); //Mark packet ready NOW
        
        NutMutexUnlock(&mutex_rfc_send_status_queue);
        return 0;
    } else {
        //Happens only, if the queue contains messages from fully
        //processed packets (failed, delivered). 
        ROUTING_DEBUG("\nW: (rfc_send_status_queue_add_src) Queue full!\n");
        
        NutMutexUnlock(&mutex_rfc_send_status_queue);
        return -1;
    }
}

static inline void rfc_send_status_queue_set_status(int handle, u_char status)
{
    NutMutexLock(&mutex_rfc_send_status_queue);
    rfc_send_status_queue[handle].status=status;
    NutMutexUnlock(&mutex_rfc_send_status_queue);
}

static inline void rfc_send_status_queue_set_nextshd(int handle, long nextshd)
{
    NutMutexLock(&mutex_rfc_send_status_queue);
    rfc_send_status_queue[handle].nextshd = nextshd;
    NutMutexUnlock(&mutex_rfc_send_status_queue);
}

void rfc_send_status_queue_clean(void)
{
    u_char i;
    
    //IMPROVE - Probably you just want to drop the oldest entry. Change behaviour here. 
    //Currently we simply drop all finished entries.
    NutMutexLock(&mutex_rfc_send_status_queue);
    for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
        if(rfc_send_status_queue[i].status == RFC_STATUS_FAILED ||
           rfc_send_status_queue[i].status == RFC_DESTINATION_UNREACHABLE ||
           rfc_send_status_queue[i].status == RFC_STATUS_SENT ||
           rfc_send_status_queue[i].status == RFC_STATUS_SENT_TO_NEXT_HOP ||
           rfc_send_status_queue[i].status == RFC_STATUS_REJECTED) {
            rfc_send_status_queue[i].status = RFC_STATUS_UNDEF;
            rfc_send_status_queue[i].retry = 0;
            rfc_send_status_queue[i].src = 0;
            rfc_send_status_queue[i].seqno = 0;
            rfc_send_status_queue[i].nextshd = 0;
        }
    }
    NutMutexUnlock(&mutex_rfc_send_status_queue);
}

int rfc_send_status_queue_get_handle(u_short src, u_short seqno)
{
    int i;
    
    NutMutexLock(&mutex_rfc_send_status_queue);
    
    for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) {
        if(rfc_send_status_queue[i].src == src && rfc_send_status_queue[i].seqno == seqno) {
            NutMutexUnlock(&mutex_rfc_send_status_queue);
            return i;
        }
    }
    
    //Error, seqno not in list
    ROUTING_DEBUG("\n\E: (rfc_send_status_queue_get_handle) Handle not in list!\n");
    
    NutMutexUnlock(&mutex_rfc_send_status_queue);
    
    return -1;
}

static inline int rfc_send_status_queue_inc_retry(int handle)
{
    int retry_count;
    
    NutMutexLock(&mutex_rfc_send_status_queue);
    retry_count = ++rfc_send_status_queue[handle].retry;
    NutMutexUnlock(&mutex_rfc_send_status_queue); 
    
    return retry_count;  
}

static inline int rfc_send_status_queue_update(int handle, u_char status, long nextshd)
{
    int retry_count;
    
    //Get lock
    NutMutexLock(&mutex_rfc_send_status_queue);
    
    //Set status
    rfc_send_status_queue[handle].status=status;
    
    //Increment retry
    retry_count = ++rfc_send_status_queue[handle].retry;
    
    //Set next schedule
    rfc_send_status_queue[handle].nextshd = nextshd;
    
    //Release lock
    NutMutexUnlock(&mutex_rfc_send_status_queue);
    
    //Return retry count
    return retry_count;
}



static inline void rfc_send_queue_del(int position)
{
    NutMutexLock(&mutex_rfc_send_queue);

    //Free packet
    if(rfc_send_queue[position] != NULL) {

        // free packet mem
        free(rfc_send_queue[position]);

        // shift entries
        for ( ; position < RFC_SEND_QUEUE_SIZE-1; position++) {
            rfc_send_queue[position] = rfc_send_queue[position+1];
        }
        // last entry is null
        rfc_send_queue[RFC_SEND_QUEUE_SIZE-1] = NULL;
        
        // queue size
        rfc_send_queue_count--;
        
    } else {
        ROUTING_DEBUG("\nW: (rfc_send_queue_del) Tried to free slot in send queue that was NULL!\n");
    }
    
    NutMutexUnlock(&mutex_rfc_send_queue);
}


/************************************************************************************
* Node and Status Information
************************************************************************************/

u_short rfc_get_address(void) 
{
    return routing->get_address();
}

void rfc_enable_led(int enable_led)
{
    rfc_use_led = (enable_led) ? 1 : 0;
}

void rfc_enable_ack(char enable_ack) 
{
    rfc_use_ack = (enable_ack) ? 1 : 0;
}

/************************************************************************************
* Initialization
************************************************************************************/

int rfc_init(routing_interface_t *routing_interface) 
{
    int i;
    
    ROUTING_DEBUG("\n\n<-------- Initializing RFC... -------->\n\n ");
    
    if(routing_interface == NULL) {
        ROUTING_DEBUG("\nE: (rfc_init) Invalid routing interface!\n");
        return -1;
    } else {
        routing = routing_interface;
    }
    
    //Initialize Mutexes
    NutMutexInit(&mutex_rfc_send_queue);
    NutMutexInit(&mutex_rfc_send_status_queue);
    NutMutexInit(&mutex_rfc_recv_queue_count);
    
    //Initialize Port listeners
    for(i = 0; i < RFC_PACKET_HANDLER_QUEUE_SIZE; i++) {
        rfc_packet_handlers[i].port = 0;
        rfc_packet_handlers[i].handler = NULL;
    }
    
    //Initialize receive queue
    rfc_recv_queue_count = 0;
    for(i = 0; i < RFC_RECV_QUEUE_SIZE; i++) {
        rfc_recv_queue[i].status = FREE;
        rfc_recv_queue[i].pkt = NULL;
    }
    
    //Initialize send queue
    rfc_send_queue_count = 0;
    for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++)
        rfc_send_queue[i] = NULL;
    
    //Initialize send status queue
    for(i = 0; i < RFC_SEND_QUEUE_SIZE; i++) 
        rfc_send_status_queue[i].status = RFC_STATUS_UNDEF;
    
    //Initialize the Sequence no.
    rfc_seqno = 0;
    
    //Initialize Random Number generator 
    //IMPROVE - better initialization
    srand((unsigned int)NutGetMillis());
    
    //Register Handler
    ccc_register_packet_handler(RFC_TYPE, rfc_receive_handler);
    routing->reg_msg_handler(routing_message_handler);
    
    //Start threads
    if( 0 == NutThreadCreate("rfc_send", rfc_send_thread, 0 , 1024)) { //IMPROVE - Decide correct stack size
                                                                       //Creating Thread failed, RFC will not operate
        ROUTING_DEBUG("\nE: (rfc_init) Creating rfc_send_thread failed!\n");
        return -1;
    }
    if( 0 == NutThreadCreate("rfc_recv", rfc_recv_thread, 0 , 1024)) { //IMPROVE -  Decide correct stack size
                                                                       //Creating Thread failed, RFC will not operate
        ROUTING_DEBUG("\nE: (rfc_init) Creating rfc_recv_thread failed!\n");
        return -1;
    }
    
    return 0;
}
#endif //__BTNODE3__
