/*
 * Copyright (C) 2000-2005 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch/
 */

/**
 * \author Matthias Ringwald
 *
 * \brief Simple connection manager to establish pre-defined chain topology
 * It does not implement the abstract connection manager cm_abstract.c
 * Feel free to change this
 *
 * The connectino thread uses the random number generator.
 * We assume that it is already seeded.
 * 
 * LED CODE:
 * LED 0 (blue): Previous node has connected to me
 * LED 1 (red):  Connected to next node
 */

#include <stdlib.h>      // rand
#include <sys/heap.h>    // NutHeapAlloc
#include <sys/event.h> 
#include <sys/thread.h>
#include <sys/timer.h>
#include <led/btn-led.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <cm/cm_chain.h>

//====================================================
// DEFINES

#define CONNECTION_RETRY_PERIOD_MS 30000
#define CONNECTION_RETRY_RANDOM_MS 0x7fff

//====================================================
// TYPEDEFS

typedef struct {	      
    struct btstack *bt_stack;   // Pointer to lower level hci stack 
    bt_addr_t addr;
    bt_addr_t *chain_topo;
    u_short    chain_length;
    u_char node_no;
    bt_hci_con_handle_t next_hop_hdl;
    bt_hci_con_handle_t prev_hop_hdl;
    HANDLE connection_update;
    // HANDLE first_contact;
} chain_stack_t;

//====================================================
// GLOBALS

chain_stack_t *chain_stack;

//====================================================
// PRIVATE


/*
 * connection callback handler
 * 
 * updates state and sets LEDs
 */
static void con_cb(u_char type, 
            u_char reason, 
            bt_hci_con_handle_t con_handle, 
            void* cb_arg)
{
    if (type == BT_HCI_CONNECTION) {
		bt_addr_t new_addr;
		bt_hci_get_remote_bt_addr( chain_stack->bt_stack, con_handle, new_addr);
		u_char new_index = connect_get_position_of ( new_addr );
		if (new_index == chain_stack->node_no-1) {
			// connected by prev node
			chain_stack->prev_hop_hdl = con_handle;
            // NutEventPost( &chain_stack->first_contact);
			btn_led_set(0);
		} else {
			// whatever..
		}
    }
    
    if (type == BT_HCI_DISCONNECTION){
		// disconnected from next hop?
		if (con_handle == chain_stack->next_hop_hdl){
			btn_led_clear(1);
		} // disconnected from prev hop?
		else if (con_handle == chain_stack->prev_hop_hdl) {
			btn_led_clear(0);
		} else {
			// printf("Disconnected from UNKNOWN! handle %u\n", con_handle);
		}
    }
}

/* 
 * Link the BT nodes to form a chain.
 */
THREAD( CONNECT, arg ) {
	// printf("CONNECT thread started.\n");
    
    // if (node_no != 0) {
    //     NutEventWait( &chain_stack->first_contact, NUT_WAIT_INFINITE);
    // }
    
	while (1) {
        
        // verify connection
        if (chain_stack->next_hop_hdl != BT_HCI_HANDLE_INVALID) {
            if (bt_hci_is_con_open(chain_stack->bt_stack, chain_stack->next_hop_hdl) != 0) {
                // connection is broken
                chain_stack->next_hop_hdl = BT_HCI_HANDLE_INVALID;
                btn_led_clear(1);
            }
        }
        
        // (re-)open connection 
        if (chain_stack->next_hop_hdl == BT_HCI_HANDLE_INVALID) {
            
            bt_hci_con_handle_t con_handle;
            
            // open connection    
            // -- printf("Connecting to %02x:%02x... \n", chain_topo[node_no+1][1], chain_topo[node_no+1][0]);
            
            con_handle = bt_hci_create_connection(chain_stack->bt_stack, 
                                                  BT_HCI_SYNC, 
                                                  chain_stack->chain_topo[chain_stack->node_no+1], 
                                                  BT_HCI_PACKET_TYPE_DH3, 
                                                  0, 
                                                  0, 
                                                  BT_HCI_ROLE_SWITCH_CHANGE); 
            
            if ( (int) con_handle <= 0) {
                // printf("Connection error: code -%x. Waiting...\n", -con_handle);    
                NutSleep( CONNECTION_RETRY_PERIOD_MS + (rand() & CONNECTION_RETRY_RANDOM_MS));
            } else {
                // printf("Open connection: %d. Done.\n", con_handle); 
                chain_stack->next_hop_hdl = con_handle;
                
                // link supervision timeout to 5 seconds
                // bt_hci_write_link_supervision_timeout(chain_stack->bt_stack, BT_HCI_SYNC, con_handle, 24);
                
                // connected
                btn_led_set(1);
            }
        }
		NutSleep(1000);
	}
}

//====================================================
// PUBLIC

/**
 * get position of node with addr
 *
 * \param addr of node
 * \returns position in chain, indexed by 0
 */
u_char connect_get_position_of(bt_addr_t addr){
	u_char i;
    for (i=0; i<chain_stack->chain_length; i++)
    {
        if (BD_ADDR_CMP(chain_stack->chain_topo[i], addr))
        {
			return i;
        }       
    }
	return 0xff;
} 

/**
 * get position of own node
 *
 * \param addr of node
 * \returns position in chain, indexed by 0
 */
u_short connect_get_position(void){
    return chain_stack->node_no;
}

/**
 * Init Chain Connection Manager
 *
 * \param struct btstack
 * \param topo array of node addresses
 * \param number of nodes in chain
 *
 */

void connect_init( struct btstack *stack, bt_addr_t* topo, u_short length){
    
    chain_stack  = NutHeapAllocClear(sizeof(chain_stack_t));
    chain_stack->chain_topo   = topo;
    chain_stack->chain_length = length;
    chain_stack->bt_stack     = stack;
    bt_hci_get_local_bt_addr(stack, chain_stack->addr);
	chain_stack->node_no = connect_get_position_of(chain_stack->addr);
    
    // printf("Node %0x's number is %u/%u.\n", my_addr[0], node_no, chain_length);
	chain_stack->next_hop_hdl = BT_HCI_HANDLE_INVALID;
    chain_stack->prev_hop_hdl = BT_HCI_HANDLE_INVALID;
	bt_hci_register_con_table_cb(stack, &con_cb, NULL);
    
    // last node has no one to connect to
    if (chain_stack->node_no != length-1)  {
        NutThreadCreate("connect", CONNECT, 0, 1024);
    }
}

