/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch/
 *
 */
 
/** 
 * @file bt_hci_transport_uart.h
 *
 * @brief UART specific function
 * This file defines the get_pkt and send_pkt functions for the uart-layer
 * See hci_transport_usb.h for the usb packet-sending.\n
 *
 * If you want to write your own transport-layer you have to implement
 * the three functions in this file, everything else is your choice.\n
 * In the btstack, the transport pointer is for the transport-layer
 * specific data.\n
 * You can save there anything you want?
 *
 * \author Mathias Payer <payerm@student.ethz.ch>
 * \date 2004/04/14 19:44:09 
 *
 */

#include <bt/bt_defs.h>
#include <sys/device.h>
#include <stdio.h>

// bt hci uart transport configuration
#include <arch/btn-hardware.h>


/*******************************************************************************
 * bt_hci_initdev
 ******************************************************************************/
/**
 * @brief Initializes the given uart.
 * 
 * Initializes the uart, given by uart_dev and sets the baudrate initially to
 * BT_UART_INITIAL_SPEED (->bt_defs.h).
 *
 *
 * @return O, or <0 on (not recoverable) error.
 *
 * @see  bt_hci_lowelevel_get_baudrate(), bt_hci_lowlevel_set_baudrate()
 */
int _bt_hci_initdev(struct btstack *stack);


/*******************************************************************************
 * _bt_hci_lowlevel_set_baudrate
 ******************************************************************************/
/** @brief Sets the baudrate of the host
 * 
 * Sets the baudrate of the uart (not the bt controller!) to the given value.\n
 * This function should only be called by internal functions like
 * bt_hci_set_baudrate() !
 *
 * @param uart The uart to change.
 * @param baud New baudrate
 *
 * @return O, or -1 on error.
 *
 * @see  bt_hci_set_baudrate()
 */
int _bt_hci_lowlevel_set_baudrate(int uart, u_long baud);


/*******************************************************************************
 * bt_hci_send_pkt
 ******************************************************************************/
/**
 * @brief Sends a packet to the uart
 * 
 * Sends a packet to the uart, the sender has to check if the bt-controller ist
 * still able to receive packets, else the sender has to block (wait)
 *
 * @param stack the BT-stack to use
 * @param *pktptr Pointer to a hci packet
 *
 * @return O, or -1 on error.
 *
 */
int _bt_hci_send_pkt(struct btstack *stack, u_char *pktptr);


/*******************************************************************************
 * bt_hci_get_pkt
 ******************************************************************************/
/**
 * @brief Gets a pkt from the uart.
 * This function has to be polled!
 * Here a thread blocks until a pkt is read from the uart.
 *
 * @param stack the BT-stack to use
 * @param len Pointer to the variable where the size of the data read stored
 * @param t_arrive Pointer to the variable where the arrival time of the packet
 * is stored
 *
 * @return Pointer to the pkt or O on error.
 *
 */
u_char *_bt_hci_get_pkt(struct btstack * stack, u_short* len, u_long* t_arrive);

/**
 * low-level functions for bt_acl
 *
 * To construct an hci packet on the fly, call aquire_lock, send parts with lowlevel_write
 * and release_lock
 */
void _bt_hci_aquire_write_lock( struct btstack * stack );
void _bt_hci_relase_write_lock( struct btstack * stack );
int _bt_hci_lowlevel_write( struct btstack * stack, u_char * buffer, u_short len );


