/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file l2cap_cl.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Connection-less data protocol API.
 * 
 * Provides functions to send connection-less data to directely
 * connected neighbors.
 * 
 * To receive l2cap connection-less data, an application has
 * to register its service routine at the protocol/service multiplexor,
 * together with a "PSM" - a unique number identifying the service.
 * Connection-less data can then be sent to a specific service
 * by sending it to the corresponding "PSM". Refer to #l2cap_cl_send
 * and bt_psm.h for further details.
 * 
 * \see
 * bt_psm.h \n
 * bt_l2cap.h \n
 * mhop_cl.h
 * 
 * Note: This is not an implementation of the l2cap connection-less
 * protocol as defined in the Bluetooth specification, although the packet
 * format used is equivalent to the definition of l2cap
 * connection-less packets (see note below).
 */
#ifndef _L2CAP_CL_H_
#define _L2CAP_CL_H_

#include <sys/types.h>
#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_acl_pkt_buf.h>
#include <bt/bt_psm.h>

/**
 * \name Errors
 */
//@{
/**
 * The payload of the l2cap connection-less data packet to send is too large
 */
#define L2CAP_CL_ERR_PAYLOAD_OVERFLOW               -1
//@}

/**
 * \brief Size of the PSM field.
 */
#define L2CAP_CL_PSM_FIELD_SIZE                    2


/**
 * \name Sending L2CAP connection-less Data
 */
//@{

/**
 * \brief Sends a connection-less l2cap data packet.
 * 
 * Sends connection-less l2cap data of length \p len over connection handle
 * \p con_handle to the psm \p psm.
 * 
 * Note: The maximal amount of data that can be sent is obtained by calling
 * #l2cap_cl_max_payload.
 * 
 * \param data pointer to the data to be sent
 * \param len amount of data to be sent
 * \param hdl connection handle the data shall be sent to
 * \param psm PSM the data shall be sent to
 * (the PSM identifies the service the data will be delivered to
 * on the receiving side (see #bt_psm_service_register).
 * 
 * \returns 0 or possible errors:
 * - #BT_ERR_BT_MODULE\n
 * - #BT_ERR_NO_CON\n
 * - #L2CAP_CL_ERR_ACL_PAYLOAD_OVERFLOW
 * 
 * Example:\n
\verbatim
u_char data[12];
strcpy(data, "hello world");
l2cap_cl_send(data, <con_handle>, <psm>, 12);
\endverbatim
 *
 * \see bt_psm.h
 */
long l2cap_cl_send(u_char* data,
                    u_short len,
                    bt_hci_con_handle_t hdl,
                    u_short psm);

/**
 * \brief Writes connection-less data to the bluetooth uart.
 * 
 * Writes l2cap connection-less data directely to the bt uart iff
 * unique write access has been aquired before
 * (i.e. #l2cap_cl_acquire_write_lock has been called).
 * 
 * Note: Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \param data Data to write
 * \param len Size of the data to write
 * 
 * \see
 * #l2cap_cl_acquire_write_lock \n
 * #l2cap_cl_release_write_lock
 */
inline long l2cap_cl_lowlevel_write(u_char* data, u_short len);

/**
 * \brief Acquires lock on bt-uart for sending l2cap connection-less data.
 * 
 * Locks the bluetooth uart for writing l2cap connection-less data.
 * The lock is not released until #l2cap_cl_release_write_lock has
 * been called.
 * 
 * Note: Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \param hdl Connection handle of the connected neighbor device the data
 * shall be sent to.
 * \param psm PSM the data will be sent to on the receiving side.
 * \param len Amount of data to be sent.
 * 
 * \returns 0 or possible errors:
 * 
 * \see
 * #l2cap_cl_lowlevel_write \n
 * #l2cap_cl_release_write_lock
 */
long l2cap_cl_acquire_write_lock(bt_hci_con_handle_t hdl,
                              u_short psm,
                              u_short len);

/**
 * \brief Releases unique access for sending l2cap connection-less data.
 * 
 * Releases the writing lock on the bluetooth controller triggered by
 * #l2cap_cl_acquire_write_lock.
 * 
 * Note: Applications typically do not call this function, but make use
 * of #l2cap_cl_send instead.
 * 
 * \see
 * #l2cap_cl_acquire_write_lock \n
 * #l2cap_cl_lowlevel_write
 */
inline void l2cap_cl_release_write_lock(void);
//@}


/**
 * \name L2CAP callback registration
 */
//@{
/**
 * \brief To register a callback funtion handling L2CAP data packets
 * 
 * Call this function to register a higher layer service (i.e. l2cap)
 * which handles the data that is not handled by the l2cap
 * connection-less layer (i.e. l2cap packets that are not sent to
 * the connection-less channel).
 * 
 * Note: l2cap_cl and l2cap cannot run together at the moment!
 */
void l2cap_cl_register_acl_cb(struct btstack* bt_stack, HCI_ACL_CB, void* arg);
//@}


/**
 * \name Initialization
 */
//@{
/**
 * \brief Initializes the connectionless l2cap layer. 
 * 
 * Note: The bt-stack, the ACL layer, as well as the protocol/service
 * multiplexor have to be initialized before calling this function. Thus, make
 * shure that:
 * - #bt_hci_init
 * - #bt_acl_init
 * - #bt_psm_init
 * have been called.
 * 
 * \param bt_stack Pointer to the bt-stack to use
 * \param mux Pointer to the protocol/service multiplexor to use
 */
void l2cap_cl_init(struct btstack* bt_stack, bt_psm_t* mux);
//@}


/**
 * \name Packet statistics
 */
//@{
/**
 * \brief Returns the number of sent packets.
 */
inline u_long l2cap_cl_get_outgoing_traffic(void);

/**
 * \brief Returns the number of received packets.
 */
inline u_long l2cap_cl_get_incoming_traffic(void);

/**
 * \brief Returns the number of dropped packets.
 */
inline u_long l2cap_cl_get_dropped_traffic(void);

/**
 * \brief Increments drop counter.
 */
inline void l2cap_cl_increment_drop_counter(void);

//#define L2CAP_CL_DROP_STATS_MAX_ENTRIES 7
//
///**
// * \brief Returns indexed drop counter. 
// */
//inline u_short l2cap_cl_get_indexed_drop_counter(u_short index);
//
///**
// * \brief Increments indexed drop counter.
// * Increments also the above drop counter.
// */
//inline void l2cap_cl_increment_indexed_drop_counter(u_short index);
//@}

/**
 * \brief Returns a pointer to the payload of a received l2cap
 * connection-less packet.
 * 
 * \param pkt Pointer to the received packet
 * 
 * \returns pointer to the payload
 */
inline u_char* l2cap_cl_get_data_ptr(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the arrival time of the packet passed.
 */
inline u_long l2cap_cl_get_arrival_time(bt_acl_pkt_buf* pkt_buf);

/**
 * \brief Returns the max payload size of a connection-less l2cap packet.
 */
inline u_short l2cap_cl_max_payload(void);

#endif //_L2CAP_CL_H_
