/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 *
 * $Id: rfc.h,v 1.1 2007/06/08 09:51:39 bosterm Exp $
 */

/**
 * @file rfc.h
 * @author Florian Schuetz <fschuetz@ethz.ch>
 * @brief Multihop packet-communication library for the CC1000 radio.
 *
 * This multihop packet-communiction library allows nodes to communicate,
 * even if they are not within each others communication range. RFC works
 * with any routing protocol that provides its functionality through the
 * CC1000 routing interface.
 *
 * If an application wants to send and receive packets, it should register
 * a packet handler and an error handler with RFC. To do this, it must choose
 * a free port that it attaches to. It then can begin sending packets to 
 * other nodes.
 *
 * Sending packets to other nodes is done byu calling rfc_send(...). This
 * will start a route discovery if no route to the destination is known.
 * As soon as the next hop is known, the packet is sent on it's way. 
 * Sending packets works asynchronously, which means that there can be
 * several pending packets for different destinations at the same time.
 * Packets for destinations with an available route a processed right away,
 * they do not have to wait for packets that were queued earlier to finish.
 *
 * Received packets are delivered to the port listener, if the receiving
 * node is the final destination. If the receiving node is not the final
 * destination, it is forwarded on the route.
 * 
 * In case of error, corrective action is taken. If the node that receives the
 * error is the originating node, the applicatino is informed. If it is not,
 * The packet is silently discarded. There exist hooks in the source code,
 * that allow to easily add end-to-end and intermediate node error reporting.
 * This should be implemented in the next version.
 *
 * When reading the sourcecode, you will find IMPROVE-tags. This means, that
 * the source code could/should be improved at this point in a later version.
 */

#ifndef _RFC_H_
#define _RFC_H_

#include <sys/types.h>
#include <cc/ccc.h>
#include <sys/mutex.h>
#include <led/btn-led.h>
#include "routing.h"

/* Some functionality definitions */
#define RFC_SEND_QUEUE_RETRY_TIMEOUT   300   ///< Minimum time in ms until send retry if send failed.
#define RFC_SEND_RETRY                   7   ///< Maximum number of retries.
#define RFC_SEND_QUEUE_SIZE             10   ///< Size of the queue for packets to be sent or forwarded.
#define RFC_RECV_QUEUE_SIZE             20   ///< Size of the queue for packets that are received and not yet processed.
#define RFC_RECV_STATUS_BUFFER_SIZE     50   ///< Size of the status buffer for received packets. 
#define RFC_PACKET_HANDLER_QUEUE_SIZE   10   ///< Number of available ports that applications can bind to.
#define RFC_RECV_STATUS_BUFFER_TIMEOUT 30000 ///< Time in ms that rfc packet information is cached to detect duplicates.

#define RFC_LED_DATA                  LED2   ///< Defines which led to use to show where data lies.

#define RFC_STATUS_UNDEF                0    ///< Undefined state
#define RFC_STATUS_SENT                 1    ///< State that marks packet as successfully delivered to destination.
#define RFC_STATUS_PENDING              2    ///< State that marks packet as waiting for processing.
#define RFC_STATUS_FAILED               3    ///< State that marks packet as dropped because of processing errors.
#define RFC_STATUS_SENT_TO_NEXT_HOP     4    ///< State that marks packet as successfully sent to next hop.
#define RFC_DESTINATION_UNREACHABLE     5    ///< State that marks a destination as unreachable (packets are dropped).
#define RFC_STATUS_WAITING              6    ///< State that marks packet as waiting for a message.
#define RFC_STATUS_READY                7    ///< State that marks packet as ready for sending.
#define RFC_STATUS_REJECTED             8    ///< State that marks packet as rejected by destination.
#define RFC_STATUS_WAIT_NEXT_HOP_ACK    9    ///< State that marks packet as waiting for an ack of next hop.

#define RFC_FLAG_NEXT_HOP_ACK           1    ///< If set, this packet must be acknowledged by next hop.
#define RFC_FLAG_ACK                    2    ///< If set, this packet must be acknowledged by it's final destination. \warning Not implemented. Probably this was just waiting for you. ;)

#define RFC_NO_ERROR                    0    ///< Standard return value for successfull processing.
#define RFC_ERROR_SEND_QUEUE_FULL  0xFFFF    ///< Error that is returned, when the send queue is full.
#define RFC_ERROR_RECV_QUEUE_FULL      -1    ///< Error that is returned, when the receive queue is full.

#define RFC_TYPE                       23    ///< Defines the CCC packet type that RFC data packets use.

// ----------------------------------------------------------------------------
// Typedefs
// ----------------------------------------------------------------------------
/**
 * Structure of an rfc  packet.
 * 
 * \warning Make sure that the data length field and the length of the actual data
 * are entered correctly.
 *
 * \warning Check the MTU with rfc_get_mtu(). Since RFC does not yet support fragmenting 
 * packets, this MUST be done at application level.
 */
typedef struct rfc_packet_st {
  u_char type;    ///< Type of the packet is always RFC_DATA_TYPE, leave blank, this is set internally.
  u_char port;    ///< Port that the packet is directed to. 
  u_short seqno;  ///< Sequence number of the RFC Packet. The sequence number uniquely identifies any data packet.
  u_short flags;  ///< Flags: A == Use Ack for this packet.
  u_short src;    ///< Source that sent the packet.
  u_short dst;    ///< Destination to which the packet is targeted.
  u_short length; ///< Denotes the payload length in bytes.
  u_char data[0]; ///< The payload in bytes. Payload should be delivered in network byte order.
} __attribute__((__packed__)) rfc_packet_t;


/**
 * Structure describing a port listener for an application
 *
 * \warning Make sure that these handler run short, since they block RFC operation and
 * might delay packet processing. In the worst case, this fills the receive buffer and 
 * incomming packets will be dropped. 
 * 
 * It is good practice to copy received data packets or error messages in a buffer and
 * process them from a separate thread.
 */
typedef struct rfc_packet_handler_st {
  void (*error_handler)(u_short, u_short);   ///< Application-handler for error message processing
  void (*packet_handler)(rfc_packet_t*);     ///< Application-handler for data packet processing
} rfc_packet_handler_t;



// ----------------------------------------------------------------------------
// global variables
// ----------------------------------------------------------------------------
extern u_long rfc_recv_count;         ///< Number of RFC packets targeted for this node
extern u_long rfc_failed_count;       ///< Number of failed RFC packets passing through node



// ----------------------------------------------------------------------------
// functions
// ----------------------------------------------------------------------------

/**
 * Initialises the RFC library.
 * 
 * This function initializes the RFC communication library and MUST be called
 * first. You must provide it a valid interface to the routing protocol you
 * want to use. This requires, that routing is already initialized and running.
 *
 * \warning Make sure you have initialized a routing protocoll before you
 * do initialize the RFC layer with this function.
 *
 * @param address [in] interface to the routing protocol that RFC should use
 * @return 0 on success, <> 0 otherwise.
 */
int rfc_init(routing_interface_t *routing_interface);

/**
 * Allocates space for a new RFC packet
 *
 * \attention Do not forget to free the space of this packet when you
 * are done. All packets that are sent by RFC will be buffered
 * there, which means that they are not affected by any change you
 * make to the original packet after they have been sent.
 *
 * @param data_len [in] length of the data to be transmitted
 * @return pointer to packet on success, 0 otherwise
 */
rfc_packet_t *new_rfc_packet(int data_len);

/**
 * Sends a Packet to the destination
 * 
 * Calling this starts sending a packet towards its destination. If a packet
 * can not be delivered, the application will be notified, if it has 
 * registired an error handler.
 *
 * \warning The end-to-end error reporting functionality is not yet implemented. 
 * However, sophisticated hooks exist so that this funtionallity can easily be added 
 * (by yourself?)
 *
 * \attention Do not forget to free the packet if you don't need it anymore. This can
 * safely be done since RFC keeps its own, local copies of packets.
 * 
 * @param dst [in] destination address
 * @param type [in] port number at which target application listens
 * @param pkt [in] pointer to the packet that should be sent
 * @return sequence no. on success, -1 otherwise.
 */
u_short rfc_send(u_short dst, u_char type, rfc_packet_t *pkt);

/**
 * Registers a port listener
 *
 * \warning Make sure that the handlers run short, since they block RFC operation and
 * might delay packet processing. In the worst case, this fills the receive buffer and 
 * incomming packets will be dropped. 
 * 
 * It is good practice to copy received data packets or error messages in a buffer and
 * process them from a separate thread.
 * 
 * @param type [in] packet type to handle
 * @param handler [in] packet handler to use
 * @return received packet.
 */
int rfc_register_packet_handler(u_char port, rfc_packet_handler_t *handler);

/**
 * Deregisters a port listener
 *
 * @param type [in] port 
 */
void rfc_deregister_packet_handler(u_char port);

/**
 * Get the maximal transfer unit
 *
 * SHOULD be used to decide maximum size of a packet that can be sent.
 *
 * @return Maximal payload size of a data packet.
 */
u_short rfc_get_mtu(void);

/**
 * Enables data ACKs
 *
 * @param enable_ack [in] 0 to disable, 1 to enable
 */
void rfc_enable_ack(char enable_ack);


/**
 * Disable RFC
 *
 * \warning Not yet implemented. Be honest, you don't want to stop RFC anyways. :)
 */
void rfc_release(void);

/**
 * Returns the configured mac address
 *
 * @return mac address of the node
 */
u_short rfc_get_address(void);


/**
 * Enables LED flashing when data is transmitted.
 *
 * The node's LEDs shines yellow, whenever a data packet
 * is being processed in the node. Whenever a delivery fails, 
 * the red LED flashes.
 *
 * @param enable_led [in] 0 to disable, 1 to enable.
 */
void rfc_enable_led(int enable_led);

#endif //_RFC_H_
