/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 * 
 * $Id: routing.h,v 1.1 2007/06/08 09:51:39 bosterm Exp $
 */

/**
 * @file routing.h
 * @author Florian Schuetz <fschuetz@ethz.ch>
 * @brief Generic Chipcon CC1000 Routing-Interface
 */

#ifndef _ROUTING_H_
#define _ROUTING_H_

#include <sys/types.h>
#include <sys/mutex.h>


/*
 * Routing messages used for status reporting
 */
#define PENDING                  1 ///< State for packet that is waiting for processing.
#define DESTINATION_UNREACHABLE  2 ///< Message, which means requested destination is unreachable.
#define DISCOVERY_STARTED        3 ///< Message, which means that a route discovery was started.
#define DISCOVERY_FAILED         4 ///< Message, which means that a route discovery failed.
#define DISCOVERY_IN_PROGRESS    7 ///< Message, which means that a route discovery for a given destination is already in progress.
#define ROUTE_AVAILABLE          5 ///< Message, which means that a route discovery was successfull and a route for a given destination is available.
#define BROKEN_LINK              6 ///< Message, which reports a broken link to a next hop on a path to a destination.
#define BROKEN_ROUTE             7 ///< Message, which reports a route that is no longer available but other nodes still use.

/** 
 * Defines the message handler, through wich the routing protocol may send
 * Messages to the RoutingForChipcon layer.
 */
typedef void (*routing_message_handler_t)(u_short destination, u_short message);


/**
 * Structure describing the Routing-Interface.
 *
 * This structure enables the RoutingForChipcon-layer to communicate
 * with an arbitrary routing protocol, that exports this interface.
 */
typedef struct routing_interface_st {
  
  /** 
   * SHOULD be used to determine nodes CC1000 interface address. 
   *
   * @return Configured mac address.
   */
  u_short (*get_address)(void);
    
  /**
   * MUST be used to determine the next hop towards a destination.
   *
   * @param dst [in] destination address
   * @return Next hop address on success, 0 if not known.
   */
  u_short (*get_next_hop)(u_short *dst);

  /**
   * Starts a route discovery for a given destionation. 
   *
   * If there is already a discovery in progress, the routing protocol
   * SHOULD ignore this call.
   *
   * @param dst [in] destination address
   * @return DISCOVERY_STARTED if a discovery was started, 
   *         ROUTE_AVAILABLE if route is already available, 
   *         DISCOVERY_IN_PROGRESS if a discovery is already in progress.
   */
  u_short (*discover)(u_short *dst);
  
  /**
   * Sends a message to the routing protocol
   *
   * Legitimate Messages are: BROKEN_LINK.
   *
   * @param msg [in] message
   * @param param [in] parameter for the corresponding message
   */
  void (*send_msg)(u_short msg, u_short param);  
  
  /**
   * MUST be used to register a handler to receive status messages from the routing
   * protocol.
   *
   * @param msg_handler [in] handler that processes messages from the routing protocol
   * @return 0 on success, -1 on failure.
   */
  int (*reg_msg_handler)(routing_message_handler_t msg_handler);


} routing_interface_t;


/*
 * define this if debug output should be printed
 */
//#define ROUTING_DEBUGMODE            ///< If defined, debug output is print to the console.
//#define ROUTING_INFOMODE             ///< If defined, additional information is print to the console. \warning Only use together with ROUTING_DEBUGMODE.

//Debugmode
#ifdef ROUTING_DEBUGMODE
MUTEX routing_Terminal;
#include <stdio.h>
#define ROUTING_DEBUG_INIT NutMutexInit(&routing_Terminal); ///< MUST be used to initialize ROUTING_DEBUG(...). \warning Always do this before using ROUTING_DEBUG(...)
#define ROUTING_DEBUG(str, ...) { NutMutexLock(&routing_Terminal); printf_P(PSTR(str), ## __VA_ARGS__); fflush(stdout); NutMutexUnlock(&routing_Terminal); } ///< Use this to print debug output.
#else
#define ROUTING_DEBUG_INIT {}      ///< MUST be used to initialize ROUTING_DEBUG(...). \warning Always do this before using ROUTING_DEBUG(...)
#define ROUTING_DEBUG(str, ...) {} ///< Use this to print debug output.
#endif


#endif /* _ROUTING_H_ */
