#ifndef _SLOG_H_
#define _SLOG_H_

#include <bt/bt_psm.h>
#include <io.h>

/**
 * \brief Sends logged messages to a target device.
 * 
 * \param target Bluetooth address of the target device
 * \param psm PSM the messages shall be sent to on the target device
 * \param nr_msgs Maximal number of messages to send
 * \param class Log class to send - pass zero to send messages of all
 * log classes
 * \param mask Log mask to use for the specified class
 * \param id_newer Only messages that have a higer id than id_newer are sent.
 * Use 0 to get all messages.
 */
void slog_send(bt_addr_t target, u_short psm, u_short nr_msgs, u_char class, u_char mask);


void slog_send_messages(bt_addr_t target_addr, u_short target_psm, u_short id_newer, u_short nr_msgs);


/**
 * \brief Gets remote log messages.
 * 
 * \param rem_addr Bluetooth address of the remote device
 * \param reply_psm PSM the messages shall be sent to (on the local device)
 * \param nr_msgs Maximal number of log messages to retrieve
 * \param class Log class to retrieve - pass zero to retrieve messages of
 * all log classes
 * \param mask Log mask to use for the specified class
 */
long slog_send_rqst(bt_addr_t rem_addr, u_short reply_psm, u_short nr_msgs, u_char class, u_char mask);

/**
 * \brief Specifies where verbose log messages are sent to.
 * 
 * After calling this function, log messages added to the log buffer are
 * immediately sent to the device specified by \p target.
 * 
 * \param target Bluetooth address of the device the messages shall be sent to
 * \param psm PSM the messages shall be sent to on the target device
 */
void slog_verbose_target_set(bt_addr_t target, u_short psm);

/**
 * \brief Returns the Bluetooth address of the device verbose log messages are
 * currently sent to.
 */
inline u_char* slog_verbose_target_get(void);

/**
 * \brief Clears the send log verbosity target
 */
void slog_verbose_target_clear(void);

int slog_verbosity_on(void);

/**
 * \brief Filter for sending verbose log messages.
 * 
 * Sets the log mask for sending verbose messages (for a class given), i.e.
 * if verbose log messages of class \p class are sent to the target device
 * (registered by calling #slog_verbose_target_set) or not is defined by the
 * log mask \p mask passed.
 * 
 * \param class Log class to set
 * \param mask Log mask
 */
void slog_verbosity_set(u_char class, u_char mask);

/**
 * \brief Returns the verbosity log mask of the class specified.
 */
u_char slog_verbosity_get(u_char class);


/**
 * \brief Initializes send log functionality.
 * 
 * Note: Multi-hop transport layer has to be initialized before calling
 * this function!
 * 
 * \param psmux Pointer to the protocol/service multiplexor to use.
 * \param rqst_psm PSM to use.
 */
void slog_init(bt_psm_t* psmux, u_short rqst_psm);

#endif //_SLOG_H_
