/*
 * Copyright (C) 2000-2005 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file wsnlab/wsnlab.c
 *
 * \date 16.05.2006
 *
 * \author bo
 *
 * \brief Helper library for the WSN-Lab
 *
 */
 
#include <stdio.h>
#include <string.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <hardware/btn-hardware.h>
#include <led/btn-led.h>
#include <cc/ccc.h>
#include <wsnlab/wsnlab.h> 
#include <btsense/btsense.h> 
#include <eepromdb/btn-eepromdb.h>
#include <terminal/btn-terminal.h>
#include <terminal/btn-cmds.h>
#include <dev/usartavr.h>

// ----------------------------------------------------------------------------
// #define constants
// ----------------------------------------------------------------------------

#define BYTES_PER_LINE 16

// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

/**
 * Prints a string representation of the data to stdout (like a hex editor)
 */
void print_data( char* data, u_short length )
{
	u_short idx = 0, jdx = 0;
	u_short numLines = (length / BYTES_PER_LINE) + (length % BYTES_PER_LINE > 0 ? 1 : 0);
	
	for ( idx = 0; idx < numLines; idx++ )
	{		
		printf( "%04u   ", idx * BYTES_PER_LINE );
	
		for ( jdx = idx*BYTES_PER_LINE; jdx < (idx+1)*BYTES_PER_LINE; jdx++ )
		{
			if ( jdx-idx*BYTES_PER_LINE == BYTES_PER_LINE / 2 )
			{
				printf( " " );
			}				
			
			if ( jdx < length )
			{
				printf( "%02X ", data[jdx] );
			}
			else
			{
				printf( "   " );
			}			
		}
		
		printf( "  " );
		for ( jdx = idx*BYTES_PER_LINE; jdx < (idx+1)*BYTES_PER_LINE; jdx++ )
		{
			if ( jdx < length ) 
			{			
				if ( data[jdx] < 32 )
				{
					// do not print special characters
					printf( "." );
				}
				else
				{
					printf( "%c", data[jdx] );
				}
			}
		}
		printf( "\n" );	
	}
	
	printf( "Length of data block: %u bytes\n", length );
}

void print_packet_short( ccc_packet_t* pkt )
{
	if ( pkt != NULL )
	{
		printf( "%04X | %04X |   %02X |   %4u | %s\n",
		pkt->dst, pkt->src, pkt->type, pkt->length, pkt->data );
	}
	else
	{
		printf( " *** NO PACKET ***\n" );
	}
}

inline u_short pow2(u_short e)
{
    return 1 << e;
} 

// FIXME: This could be done with a single hw write operation
void led_set_binary(u_short i)
{
	u_short e=3;
	btn_led_clear(0);
	btn_led_clear(1);
	btn_led_clear(2);
	btn_led_clear(3);
	while (i)
	{
		while(i < pow2(e)) e--;
		btn_led_set(e);
		i -= pow2(e);
	}
}

/**
 * Helper function for sending packets via the CCC. 
 */
int ccc_send_packet( u_short dest, u_char type, u_char* data, u_short length )
{
	int res = 0;
	ccc_packet_t *pkt = new_ccc_packet( length );
	
	pkt->length = length;
	memcpy( pkt->data, data, length );
	res = ccc_send( dest, type, pkt );
	
	free( pkt );
	
	return res;
}

void beep()
{
	btsense_sound(440);
    NutSleep(1000);
    btsense_sound(0);
}


u_short get_MAC_address_from_EEPROM(void)
{
	u_long address;
		
    if ( btn_eepromdb_get( FOURCC('M','A','C',' '), 
    					   BTN_EEPROMDB_INT_SIZE, 
    					   &address ) == 0 )
	{
		return 0;
	}
	
	return (u_short) address;
}

/*
 * reads/writes the MAC-address to the internal EEPROM
 */
void address_cmd(char* arg)
{
	u_long address;	
	
	// trim left
	while ( *arg == ' ' || *arg == '\t' )
	{
		arg++;
	}
	
	// check if there is a parameter
	if ( strlen( arg ) > 0 )
	{
		// convert string to integer
		if ( sscanf( arg, "%lx", &address ) <= 0 )
		{
			printf( "error: positive number required!\n" );
			return;									
		}
		
		// check if address <= 0xFFFF
		if ( address > 0xFFFF )
		{
			printf( "error: address must be between 0x0000 and 0xFFFF!\n" );
			return;												
		}

		printf( "writing 0x%08lX to EEPROM...\n", address );
		
		if ( btn_eepromdb_store( FOURCC('M','A','C',' '), 4, &address) == 0 )
		{
			printf( "error while storing!\n" );
			return;
		}

		printf( "OK\n" );
	}
	else
	{
		// read out stored MAC address
	    if ( btn_eepromdb_get( FOURCC('M','A','C',' '), 4, &address ) == 0 )
		{
			printf( "Error: No MAC address found in EEPROM!\n" );
			return;
		}
		
		printf( "MAC address in EEPROM = 0x%04lX\n", address );
	}
}

void wsnlab_cmds_register_cmds(void)
{
	btn_terminal_register_cmd("address", address_cmd);    		
}


/**
 * Initialises the first UART and the terminal.
 */
void term_init(u_long baud, char* prompt) 
{
    /* initialize UART 1 */
    NutRegisterDevice(&APP_UART, 0, 0);
    stdin = fopen(APP_UART.dev_name, "r+");
    _ioctl(_fileno(stdin), UART_SETSPEED, &baud);
    btn_terminal_init(stdin, prompt);
    freopen(APP_UART.dev_name, "w", stdout);
}


